import numpy as np
import plotly.express as px
import plotly.graph_objects as go

from estimagic.benchmarking.process_benchmark_results import (
    process_benchmark_results,
)
from estimagic.config import PLOTLY_TEMPLATE
from estimagic.utilities import propose_alternatives
from estimagic.visualization.plotting_utilities import create_grid_plot, create_ind_dict


def convergence_plot(
    problems,
    results,
    *,
    problem_subset=None,
    algorithm_subset=None,
    n_cols=2,
    distance_measure="criterion",
    monotone=True,
    normalize_distance=True,
    runtime_measure="n_evaluations",
    stopping_criterion="y",
    x_precision=1e-4,
    y_precision=1e-4,
    combine_plots_in_grid=True,
    template=PLOTLY_TEMPLATE,
    palette=px.colors.qualitative.Plotly,
):
    """Plot convergence of optimizers for a set of problems.

    This creates a grid of plots, showing the convergence of the different
    algorithms on each problem. The faster a line falls, the faster the algorithm
    improved on the problem. The algorithm converged where its line reaches 0
    (if normalize_distance is True) or the horizontal blue line labeled "true solution".

    Each plot shows on the x axis the runtime_measure, which can be walltime, number
    of evaluations or number of batches. Each algorithm's convergence is a line in the
    plot. Convergence can be measured by the criterion value of the particular
    time/evaluation. The convergence can be made monotone (i.e. always taking the bast
    value so far) or normalized such that the distance from the start to the true
    solution is one.

    Args:
        problems (dict): estimagic benchmarking problems dictionary. Keys are the
            problem names. Values contain information on the problem, including the
            solution value.
        results (dict): estimagic benchmarking results dictionary. Keys are
            tuples of the form (problem, algorithm), values are dictionaries of the
            collected information on the benchmark run, including 'criterion_history'
            and 'time_history'.
        problem_subset (list, optional): List of problem names. These must be a subset
            of the keys of the problems dictionary. If provided the convergence plot is
            only created for the problems specified in this list.
        algorithm_subset (list, optional): List of algorithm names. These must be a
            subset of the keys of the optimizer_options passed to run_benchmark. If
            provided only the convergence of the given algorithms are shown.
        n_cols (int): number of columns in the plot of grids. The number
            of rows is determined automatically.
        distance_measure (str): One of "criterion", "parameter_distance".
        monotone (bool): If True the best found criterion value so far is plotted.
            If False the particular criterion evaluation of that time is used.
        normalize_distance (bool): If True the progress is scaled by the total distance
            between the start value and the optimal value, i.e. 1 means the algorithm
            is as far from the solution as the start value and 0 means the algorithm
            has reached the solution value.
        runtime_measure (str): "n_evaluations", "walltime" or "n_batches".
        stopping_criterion (str): "x_and_y", "x_or_y", "x", "y" or None. If None, no
            clipping is done.
        x_precision (float or None): how close an algorithm must have gotten to the
            true parameter values (as percent of the Euclidean distance between start
            and solution parameters) before the criterion for clipping and convergence
            is fulfilled.
        y_precision (float or None): how close an algorithm must have gotten to the
            true criterion values (as percent of the distance between start
            and solution criterion value) before the criterion for clipping and
            convergence is fulfilled.
        combine_plots_in_grid (bool): decide whether to return a one
            figure containing subplots for each factor pair or a dictionary
            of individual plots. Default True.
        template (str): The template for the figure. Default is "plotly_white".
        palette: The coloring palette for traces. Default is "qualitative.Plotly".

    Returns:
        plotly.Figure: The grid plot or dict of individual plots

    """

    df, _ = process_benchmark_results(
        problems=problems,
        results=results,
        stopping_criterion=stopping_criterion,
        x_precision=x_precision,
        y_precision=y_precision,
    )

    # handle string provision for single problems / algorithms
    if isinstance(problem_subset, str):
        problem_subset = [problem_subset]
    if isinstance(algorithm_subset, str):
        algorithm_subset = [algorithm_subset]

    _check_only_allowed_subset_provided(problem_subset, df["problem"], "problem")
    _check_only_allowed_subset_provided(algorithm_subset, df["algorithm"], "algorithm")

    if problem_subset is not None:
        df = df[df["problem"].isin(problem_subset)]
    if algorithm_subset is not None:
        df = df[df["algorithm"].isin(algorithm_subset)]

    # plot configuration
    outcome = (
        f"{'monotone_' if monotone else ''}"
        + distance_measure
        + f"{'_normalized' if normalize_distance else ''}"
    )

    remaining_problems = df["problem"].unique()
    n_rows = int(np.ceil(len(remaining_problems) / n_cols))

    # pre - style plots labels
    y_labels = {
        "criterion": "Current Function Value",
        "monotone_criterion": "Best Function Value Found So Far",
        "criterion_normalized": "Share of Function Distance to Optimum<br>"
        "Missing From Current Criterion Value",
        "monotone_criterion_normalized": "Share of Function Distance to Optimum<br>"
        "Missing From Best So Far",
        "parameter_distance": "Distance Between Current and Optimal Parameters",
        "parameter_distance_normalized": "Share of Parameter Distance to Optimum<br>"
        "Missing From Current Parameters",
        "monotone_parameter_distance_normalized": "Share of the Parameter Distance "
        "to Optimum<br> Missing From the Best Parameters So Far",
        "monotone_parameter_distance": "Distance Between the Best Parameters So Far<br>"
        "and the Optimal Parameters",
    }

    x_labels = {
        "n_evaluations": "Number of Function Evaluations",
        "walltime": "Elapsed Time",
        "n_batches": "Number of Batches",
    }

    # container for individual plots
    g_list = []
    # container for titles
    titles = []

    # creating data traces for plotting faceted/individual plots
    # dropping usage of palette for algoritms, but use the built in pallete
    for prob_name in remaining_problems:
        g_ind = []  # container for data for traces in individual plot
        to_plot = df[df["problem"] == prob_name]
        if runtime_measure == "n_batches":
            to_plot = (
                to_plot.groupby(["algorithm", runtime_measure]).min().reset_index()
            )

        for i, alg in enumerate(to_plot["algorithm"].unique()):
            temp = to_plot[to_plot["algorithm"] == alg]
            trace_1 = go.Scatter(
                x=temp[runtime_measure],
                y=temp[outcome],
                mode="lines",
                legendgroup=i,
                name=alg,
                line={"color": palette[i]},
            )
            g_ind.append(trace_1)

        if distance_measure == "criterion" and not normalize_distance:
            f_opt = problems[prob_name]["solution"]["value"]
            trace_2 = go.Scatter(
                y=[f_opt for i in to_plot[runtime_measure]],
                x=to_plot[runtime_measure],
                mode="lines",
                line={"color": palette[i + 1]},
                name="true solution",
                legendgroup=i + 1,
            )
            g_ind.append(trace_2)

        g_list.append(g_ind)
        titles.append(prob_name.replace("_", " ").title())

    xaxis_title = [x_labels[runtime_measure] for ind in range(len(g_list))]
    yaxis_title = [y_labels[outcome] for ind in range(len(g_list))]

    common_dependencies = {
        "ind_list": g_list,
        "names": titles,
        "clean_legend": True,
        "x_title": xaxis_title,
        "y_title": yaxis_title,
    }
    common_layout = {
        "template": template,
        "margin": {"l": 10, "r": 10, "t": 30, "b": 10},
    }

    # Plot with subplots
    if combine_plots_in_grid:
        g = create_grid_plot(
            rows=n_rows,
            cols=n_cols,
            **common_dependencies,
            kws={"height": 320 * n_rows, "width": 500 * n_cols, **common_layout},
        )
        out = g

    # Dictionary for individual plots
    else:
        ind_dict = create_ind_dict(
            **common_dependencies,
            kws={"height": 320, "width": 500, "title_x": 0.5, **common_layout},
        )

        out = ind_dict

    return out


def _check_only_allowed_subset_provided(subset, allowed, name):
    """Check if all entries of a proposed subset are in a Series.

    Args:
        subset (iterable or None): If None, no checks are performed. Else a ValueError
            is raised listing all entries that are not in the provided Series.
        allowed (iterable): allowed entries.
        name (str): name of the provided entries to use for the ValueError.

    Raises:
        ValueError

    """
    allowed = set(allowed)
    if subset is not None:
        missing = [entry for entry in subset if entry not in allowed]
        if missing:
            missing_msg = ""
            for entry in missing:
                proposed = propose_alternatives(entry, allowed)
                missing_msg += f"Invalid {name}: {entry}. Did you mean {proposed}?\n"
            raise ValueError(missing_msg)



import sys
from traceback import format_exception


class EstimagicError(Exception):
    """Base exception for estimagic which should be inherited by all exceptions."""


class TableExistsError(EstimagicError):
    """Exception for database tables that should not exist but do."""


class InvalidFunctionError(EstimagicError):
    """Exception for invalid user provided functions.

    This includes user functions that do not comply with interfaces, raise errors or
    produce NaNs.

    """


class UserFunctionRuntimeError(EstimagicError):
    """Exception that is raised when user provided functions raise errors."""


class InvalidKwargsError(EstimagicError):
    """Exception for invalid user provided keyword arguments."""


class InvalidParamsError(EstimagicError):
    """Exception for invalid user provided parameters."""


class InvalidConstraintError(EstimagicError):
    """Exception for invalid user provided constraints."""


class InvalidBoundsError(EstimagicError):
    """Exception for invalid user provided bounds."""


class NotInstalledError(EstimagicError):
    """Exception when optional dependencies are needed but not installed."""


class NotAvailableError(EstimagicError):
    """Exception when something is not available, e.g. because a calculation failed."""


class StopOptimizationError(EstimagicError):
    def __init__(self, message, current_status):
        super().__init__(message)
        self.message = message
        self.current_status = current_status

    def __reduce__(self):
        """Taken from here: https://tinyurl.com/y6eeys2f."""
        return (StopOptimizationError, (self.message, self.current_status))


def get_traceback():
    tb = format_exception(*sys.exc_info())
    if isinstance(tb, list):
        tb = "".join(tb)
    return tb


INVALID_INFERENCE_MSG = (
    "Taking the inverse of the information matrix failed. Only ever use this "
    "covariance matrix or standard errors based on it for diagnostic purposes, not for "
    "drawing conclusions."
)


INVALID_SENSITIVITY_MSG = (
    "Taking inverse failed during the calculation of sensitvity measures. Interpret "
    "them with caution."
)



import numpy as np
from scipy import stats
from scipy.linalg import pinv
from scipy.ndimage import convolve1d


def richardson_extrapolation(sequence, steps, method="central", num_terms=None):
    """Apply Richardson extrapolation to sequence.

    Suppose you have a series expansion

        L = g(h) + a0*(h**p_0) + a1*(h**p_1) + a2*(h**p_2) + ... ,

    where p_i = order + exponentiation_step * i  and g(h) -> L as h -> 0, but g(0) != L.

    For ``method``='central', that is, for a sequence resulting from a central
    differences derivative approximation, we get ``order`` = 2 and
    ``exponentiation_step`` = 2, which would result in

        L = g(h) + a0*(h**2) + a1*(h**4) + a2*(h**6) + ...,

    where g(h) := [f(x + h) - f(x - h)] / 2h and f the function of interest. See
    function ``_get_order_and_exponentiation_step`` for more details.

    If we evaluate the right hand side for different stepsizes h we can fit a polynomial
    to that sequence of approximations and use the estimated intercept as a better
    approximation for L. Further, we can compute estimation errors of our approximation.


    Args:
        sequence (np.ndarray): The sequence of which we want to approximate the limit.
            Has dimension (k x n x m), where k denotes the number of sequence elements
            and an element ``sequence[l, :, :]`` denotes the (n x m) dimensional element

        steps (namedtuple): Namedtuple with the field names pos and neg. Each field
            contains a numpy array of shape (n_steps, len(x)) with the steps in
            the corresponding direction. The steps are always symmetric, in the sense
            that steps.neg[i, j] = - steps.pos[i, j] unless one of them is NaN.

        method (str): One of ["central", "forward", "backward"], default "central".

        num_terms (int): Number of terms needed to construct one estimate. Default is
            ``steps.shape[0] - 1``.

    Returns:
        limit (np.ndarray): The refined limit.
        error (np.ndarray): The error approximation of ``limit``.

    """
    seq_len = sequence.shape[0]
    steps = steps.pos
    n_steps = steps.shape[0]
    num_terms = n_steps if num_terms is None else num_terms

    assert (
        seq_len == n_steps
    ), "Length of ``steps`` must coincide with length of ``sequence``."
    assert num_terms > 0, "``num_terms`` must be greater than zero."
    assert (
        seq_len - 1 >= num_terms
    ), "``num_terms`` cannot be greater than ``seq_len`` - 1."

    step_ratio = _compute_step_ratio(steps)
    order, exponentiation_step = _get_order_and_exponentiation_step(method)

    richardson_coef = _richardson_coefficients(
        num_terms,
        step_ratio,
        exponentiation_step,
        order,
    )

    new_sequence = convolve1d(
        input=sequence, weights=richardson_coef[::-1], axis=0, origin=num_terms // 2
    )

    m = seq_len - num_terms
    mm = m + 1 if num_terms >= 2 else seq_len

    abserr = _estimate_error(new_sequence[:mm], sequence, richardson_coef)

    limit = new_sequence[:m]
    error = abserr[:m]

    return limit, error


def _richardson_coefficients(num_terms, step_ratio, exponentiation_step, order):
    """Compute Richardson coefficients.

    Let e := ``exponentiation_step``, r := ``step_ratio``, o := ``order`` and
    n := ``num_terms``. We build a matrix of the form

            [[1      1                  ...         1                ],
             [1    1/(s)**(2*o)         ...  1/(s)**(2*(o+n))        ],
        R =  [1    1/(s**2)**(2*o)      ...  1/(s**2)**(2*(o+n))     ],
             [...                       ...        ...               ],
             [1    1/(s**(n+1))**(2*o)  ...  1/(s**(n+1))**(2*(o+n)) ]]

    which is the weighting matrix in equation 24 in https://tinyurl.com/ybtfj4pm.
    We then return the first row of R^{-1} as the coefficients, as can be seen in
    equation 25 in https://tinyurl.com/ybtfj4pm.


    Args:
        num_terms (int): Number of terms needed to construct one estimate. Default is
            ``steps.shape[0] - 1``.

        step_ratio (float): Ratio between two consecutive steps. Order is chosen such
            that ``step_ratio`` >= 1.

        exponentiation_step (int): Step representing the growth of the exponent in
            the series expansions of the limit.
            For central difference derivative approximation ``exponentiation_step`` = 2.

        order (int): Initial order of the approximation error of sequence elements.
            For central difference derivative approximation ``order`` = 2.

    Returns:
        coef (np.ndarray): Richardson coefficients, array has length num_terms + 1.

    Example:
    >>> import numpy as np
    >>> num_terms = 2
    >>> step_ratio = 2.
    >>> exponentiation_step = 2
    >>> order = 2
    >>> _richardson_coefficients(num_terms, step_ratio, exponentiation_step, order)
    array([ 0.02222222, -0.44444444,  1.42222222])

    """
    rows, cols = np.ogrid[: num_terms + 1, :num_terms]

    coef_mat = np.ones((num_terms + 1, num_terms + 1))
    coef_mat[:, 1:] = (1.0 / step_ratio) ** (
        rows @ (exponentiation_step * cols + order)
    )

    coef = pinv(coef_mat)[0]
    return coef


def _estimate_error(new_seq, old_seq, richardson_coef):
    """Estimate error of multiple Richardson limit approximation.

    Args:
        new_seq (np.ndarray): Multiple estimates of the limit of ``old_seq``. The last
            two dimensions coincide with those of ``old_seq``. The first dimensions
            denotes the number of different estimates.

        old_seq (np.ndarray): The sequence of which we want to approximate the limit.
            Has dimension (k x n x m), where k denotes the number of sequence elements
            and an element ``sequence[l, :, :]`` denotes the (n x m) dimensional element

        richardson_coef (np.ndarray): Richardson coefficients. See function
            ``_richardson_coefficient`` for details.

    Returns:
        abserr (np.ndarray): The error estimate for each limit approximation in
            ``new_seq``.

    """
    eps = np.finfo(float).eps
    t_quantile = stats.t(df=1).ppf(0.975)  # 12.7062047361747 in numdifftools
    new_seq_len = new_seq.shape[0]

    unnormalized_covariance = np.sum(richardson_coef**2)
    fact = np.maximum(t_quantile * np.sqrt(unnormalized_covariance), eps * 10.0)

    if new_seq_len <= 1:
        delta = np.diff(old_seq, axis=0)
        tol = np.maximum(np.abs(old_seq[:-1]), np.abs(old_seq[1:])) * fact
        err = np.abs(delta)
        converged = err <= tol
        abserr = err[-new_seq_len:] + np.where(
            converged[-new_seq_len:],
            tol[-new_seq_len:] * 10,
            abs(new_seq - old_seq[-new_seq_len:]) * fact,
        )
    else:
        err = np.abs(np.diff(new_seq, axis=0)) * fact
        tol = np.maximum(np.abs(new_seq[1:]), np.abs(new_seq[:-1])) * eps * fact
        converged = err <= tol
        abserr = err + np.where(
            converged,
            tol * 10,
            abs(new_seq[:-1] - old_seq[-new_seq_len + 1 :]) * fact,
        )

    return abserr


def _get_order_and_exponentiation_step(method):
    """Return order and exponentiation step given ``method``.

    Given ``method`` we return the initial order of the approximation error of the
    sequence under consideration (order) as well as the step size representing the
    growth of the exponent in the series expansion of the limit (exponentiation_step).
    See function ``richardson_extrapolation`` for more details.

    For different methods, different values of order and exponentiation step apply.
    Consider the following examples, where we continue the notation from function
    ``richardson_extrapolation`` and use O() to denote the Big O Laundau symbol.

    Central Differences.
        Derivative approximation via central difference is given by
            g(h) := [f(x + h) - f(x - h)] / 2h = f'(x) + r(x, h),
        where r(x, h) denotes the remainder term.

        If we expand the remainder term r(x, h) we get
            r(x, h) = a0*(h**2) + a1*(h**4) + a2*(h**6) + ...
        with a0 = f''(x) / 2!, a1 = f'''(x) / 3! etc.

        Rearanging terms we can write L := f'(x) = g(h) - r(x, h) = g(h) + O(h**2) and
        we notice that order = 2 and exponentiation_step = 2.

    Forward Differences.
        Derivative approximation via forward difference is given by
            g(h) := [f(x + h) - f(x)] / h = f'(x) + r(x, h),
            where again r(x, h) denotes the remainder term.

        If we expand the remainder term r(x, h) we get
            r(x, h) = a0*(h**1) + a1*(h**2) + a2*(h**3) + ...
        with a0 = f''(x) / 2!, a1 = f'''(x) / 3! etc.

        Rearanging terms we can write L := f'(x) = g(h) - r(x, h) = g(h) + O(h) and
        we notice that order = 1 and exponentiation_step = 1.

    Backward Differences.
        Analogous to forward differences.


    Args:
        method (str): One of ["central", "forward", "backward"], default "central".

    Returns:
        order (int): Initial order of the approximation error of sequence elements.

        exponentiation_step (int): Step representing the growth of the exponent in the
            series expansions of the limit.

    Example:
    >>> _get_order_and_exponentiation_step('central')
    (2, 2)

    """
    lookup = {
        "central": (2, 2),
        "forward": (1, 1),
        "backward": (1, 1),
    }

    order, exponentiation_step = lookup[method]
    return order, exponentiation_step


def _compute_step_ratio(steps):
    """Compute the step ratio used in producing ``steps``.

    Args:
        steps (np.ndarray): Array of shape (n_steps, len(x)) with the steps in the
            corresponding direction.

    Returns:
        step_ratio (float): The step ratio used in producing ``steps``.

    Example:
    >>> import numpy as np
    >>> steps = np.array([[2., np.nan, 2], [4, 4, 4], [8, 8, np.nan]])
    >>> _compute_step_ratio(steps)
    2.0

    """
    ratios = steps[1:, :] / steps[:-1, :]
    ratios = ratios[np.isfinite(ratios)]

    step_ratio = ratios.flat[0]
    return step_ratio



import pandas as pd
import plotly.express as px

from estimagic.benchmarking.process_benchmark_results import (
    process_benchmark_results,
)
from estimagic.config import PLOTLY_TEMPLATE


def deviation_plot(
    problems,
    results,
    *,
    runtime_measure="n_evaluations",
    distance_measure="criterion",
    monotone=True,
    template=PLOTLY_TEMPLATE,
):
    """Plot average convergence of optimizers for a set of problems.

    Returns aggregated version convergence plot, showing the convergence of the
    different algorithms, averaged over a problem set. The faster a line falls, the
    faster the algorithm improved on average.

    The x axis is the runtime_measure, which can be walltime or number of evaluations.
    The y axis is the average over the convergence measures of the problems in the set.
    Convergence can be measured by the criterion value of the particular
    time/evaluation. The convergence can be made monotone by always taking the
    best  value.

    Args:
        problems (dict): estimagic benchmarking problems dictionary. Keys are the
            problem names. Values contain information on the problem, including the
            solution value.
        results (dict): estimagic benchmarking results dictionary. Keys are
            tuples of the form (problem, algorithm), values are dictionaries of the
            collected information on the benchmark run, including 'criterion_history'
            and 'time_history'.
        runtime_measure (str): One of "n_evaluations", "n_batches".
        distance_measure (str): One of "criterion", "parameter_distance".
        monotone (bool): If True the best found criterion value so far is plotted.
            If False the particular criterion evaluation of that time is used.
        template (str): The template for the figure. Default is "plotly_white".

    Returns:
        plotly.Figure

    """
    df, _ = process_benchmark_results(
        problems=problems,
        results=results,
        stopping_criterion="y",
        x_precision=1e-6,
        y_precision=1e-6,
    )

    outcome = f"{'monotone_' if monotone else ''}" + distance_measure + "_normalized"
    deviations = (
        df.groupby(["problem", "algorithm", runtime_measure])
        .min()[outcome]
        .reindex(
            pd.MultiIndex.from_product(
                [
                    df["problem"].unique(),
                    df["algorithm"].unique(),
                    range(df[runtime_measure].min(), df[runtime_measure].max() + 1),
                ],
                names=["problem", "algorithm", runtime_measure],
            )
        )
        .ffill()
        .reset_index()
    )
    average_deviations = (
        deviations.groupby(["algorithm", runtime_measure])
        .mean(numeric_only=True)[outcome]
        .reset_index()
    )
    fig = px.line(average_deviations, x=runtime_measure, y=outcome, color="algorithm")

    y_labels = {
        "criterion_normalized": "Share of Function Distance to Optimum<br>"
        "Missing From Current Criterion Value",
        "monotone_criterion_normalized": "Share of Function Distance to Optimum<br>"
        "Missing From Best So Far",
        "parameter_distance_normalized": "Share of Parameter Distance to Optimum<br>"
        "Missing From Current Parameters",
        "monotone_parameter_distance_normalized": "Share of the Parameter Distance "
        "to Optimum<br> Missing From the Best Parameters So Far",
    }
    x_labels = {
        "n_evaluations": "Numver of Function Evaluations",
        "n_batches": "Number of Batches",
    }
    fig.update_layout(
        xaxis_title=x_labels[runtime_measure],
        yaxis_title=y_labels[outcome],
        title=None,
        height=300,
        width=500,
        margin={"l": 10, "r": 10, "t": 30, "b": 10},
        template=template,
    )

    return fig



import functools
import warnings
from functools import partial

import numpy as np

from estimagic.batch_evaluators import process_batch_evaluator
from estimagic.logging.read_from_database import (
    list_of_dicts_to_dict_of_lists,
)
from estimagic.logging.write_to_database import update_row
from estimagic.algorithms import ALL_ALGORITHMS
from estimagic.utilities import propose_alternatives


def process_user_algorithm(algorithm):
    """Process the user specfied algorithm.

    If the algorithm is a callable, this function just reads out the algorithm_info
    and available options. If algorithm is a string it loads the algorithm function
    from the available algorithms.

    Args:
        algorithm (str or callable): The user specified algorithm.

    Returns:
        callable: the raw internal algorithm
        AlgoInfo: Attributes of the algorithm
        set: The free arguments of the algorithm.

    """
    if isinstance(algorithm, str):
        try:
            # Use ALL_ALGORITHMS and not just AVAILABLE_ALGORITHMS such that the
            # algorithm specific error message with installation instruction will be
            # reached if an optional dependency is not installed.
            algorithm = ALL_ALGORITHMS[algorithm]
        except KeyError:
            proposed = propose_alternatives(algorithm, list(ALL_ALGORITHMS))
            raise ValueError(
                f"Invalid algorithm: {algorithm}. Did you mean {proposed}?"
            ) from None

    algo_info = algorithm._algorithm_info

    return algorithm, algo_info


def get_final_algorithm(
    raw_algorithm,
    algo_info,
    valid_kwargs,
    lower_bounds,
    upper_bounds,
    nonlinear_constraints,
    algo_options,
    logging,
    database,
    collect_history,
):
    """Get algorithm-function with partialled options.

    The resulting function only depends on ``x``,  the relevant criterion functions,
    derivatives and ``step_id``. The remaining options are partialled in.

    Moreover, we add the following capabilities over the internal algorithms:
    - log the algorithm progress in a database (if logging is True)
    - collect the history of parameters and criterion values as well as runtime and
      batch information.

    Args:
        algorithm (str or callable): String with the name of an algorithm or internal
            algorithm function.
        lower_bounds (np.ndarray): 1d numpy array with lower bounds.
        upper_bounds (np.ndarray): 1d numpy array with upper bounds.
        nonlinear_constraints (list[dict]): List of dictionaries, each containing the
            specification of a nonlinear constraint.
        algo_options (dict): Dictionary with additional keyword arguments for the
            algorithm. Entries that are not used by the algorithm are ignored with a
            warning.
        logging (bool): Whether the algorithm should do logging.
        database (DataBase): Database to which the logging should be written.

    Returns:
        callable: The algorithm.

    """
    algo_name = algo_info.name

    internal_options = _adjust_options_to_algorithm(
        algo_options=algo_options,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        nonlinear_constraints=nonlinear_constraints,
        algo_name=algo_name,
        valid_kwargs=valid_kwargs,
    )

    algorithm = partial(raw_algorithm, **internal_options)

    algorithm = _add_logging(
        algorithm,
        logging=logging,
        database=database,
    )
    is_parallel = algo_info.parallelizes
    batch_size = internal_options.get("batch_size", internal_options.get("n_cores", 1))

    if collect_history and not algo_info.disable_history:
        algorithm = _add_history_collection(algorithm, is_parallel, batch_size)

    return algorithm


def _add_logging(algorithm=None, *, logging=None, database=None):
    """Add logging of status to the algorithm."""

    def decorator_add_logging_to_algorithm(algorithm):
        @functools.wraps(algorithm)
        def wrapper_add_logging_algorithm(**kwargs):
            step_id = int(kwargs["step_id"])

            _kwargs = {k: v for k, v in kwargs.items() if k != "step_id"}

            if logging:
                update_row(
                    data={"status": "running"},
                    rowid=step_id,
                    table_name="steps",
                    database=database,
                )

            for task in ["criterion", "derivative", "criterion_and_derivative"]:
                if task in _kwargs:
                    _kwargs[task] = partial(
                        _kwargs[task], fixed_log_data={"step": step_id}
                    )

            res = algorithm(**_kwargs)

            if logging:
                update_row(
                    data={"status": "complete"},
                    rowid=step_id,
                    table_name="steps",
                    database=database,
                )

            return res

        return wrapper_add_logging_algorithm

    if callable(algorithm):
        return decorator_add_logging_to_algorithm(algorithm)
    else:
        return decorator_add_logging_to_algorithm


def _add_history_collection(algorithm, is_parallel, batch_size):
    """Add history collection to the algorithm.

    The history collection is done jointly be the internal criterion function and the
    batch evaluator. Using the batch evaluator for history collection is necessary
    for two reasons:
    1. The batch information is only known inside the batch evaluator
    2. The normal approach of appending information to a list that is partialled into
       the internal criterion function breaks down when the batch evaluator pickles
       the internal criterion function.

    We make sure that optimizers that do some function evaluations via the batch
    evaluator and others directly are handled correctly.

    The interplay between the two methods for history collections is as follows:
    - If the function is called directly, all relevant information is appended to a list
      that is partialled into the internal criterion function.
    - If the function is called via the batch evaluator, we signal this to the internal
      criterion via the two arguments ``history_container`` (which is set to None) and
      ``return_history_entry`` (which is set to True). The returned history entries
      are then collected inside the batch evaluator and appended to the history
      container after all evaluations are done.

    Args:
        algorithm (callable): The algorithm.
        is_parallel (bool): Whether the algorithm can parallelize.

    """

    @functools.wraps(algorithm)
    def algorithm_with_history_collection(**kwargs):
        # initialize the shared history container
        container = []

        # add history collection via the internal criterion functions
        func_names = {"criterion", "derivative", "criterion_and_derivative"}
        _kwargs = kwargs.copy()
        for name in func_names:
            if name in kwargs:
                _kwargs[name] = partial(kwargs[name], history_container=container)

        # add history collection via the batch evaluator
        if is_parallel:
            raw_be = kwargs.get("batch_evaluator", "joblib")
            batch_evaluator = process_batch_evaluator(raw_be)

            _kwargs["batch_evaluator"] = _get_history_collecting_batch_evaluator(
                batch_evaluator=batch_evaluator,
                container=container,
                batch_size=batch_size,
            )

        # call the algorithm
        out = algorithm(**_kwargs)

        # add the history container to the algorithm output
        if "history" not in out:
            out["history"] = container
        else:
            out["history"] = out["history"] + container

        # process the history
        out["history"] = _process_collected_history(out["history"])

        return out

    return algorithm_with_history_collection


def _get_history_collecting_batch_evaluator(batch_evaluator, container, batch_size):
    @functools.wraps(batch_evaluator)
    def history_collecting_batch_evaluator(*args, **kwargs):
        if args:
            func = args[0]
        else:
            func = kwargs["func"]

        # find out if func is our internal criterion function. This is
        # necessary because an algorithm might use the batch evaluatior for
        # other functions as well, but for those functions we do not want to
        # (and cannot) collect a history.
        if isinstance(func, partial) and "history_container" in func.keywords:
            # partial in None as history container to disable history collection
            # via criterion function, which would not work with parallelization
            _func = partial(func, history_container=None, return_history_entry=True)

            if args:
                _args = (_func, *args[1:])
                _kwargs = kwargs
            else:
                _args = args
                _kwargs = kwargs.copy()
                _kwargs["func"] = _func

            raw_out = batch_evaluator(*_args, **_kwargs)
            out = [tup[0] for tup in raw_out]
            _hist = [tup[1] for tup in raw_out if tup[1] is not None]
            _start_batch = container[-1]["batches"] + 1 if container else 0
            _offsets = np.arange(len(_hist)).repeat(batch_size)[: len(_hist)]
            _batch_info = _offsets + _start_batch
            for batch, hist_entry in zip(_batch_info, _hist):
                hist_entry["batches"] = batch

            container.extend(_hist)

        else:
            out = batch_evaluator(*args, **kwargs)

        return out

    return history_collecting_batch_evaluator


def _process_collected_history(raw):
    history = list_of_dicts_to_dict_of_lists(raw)
    runtimes = np.array(history["runtime"])
    runtimes -= runtimes[0]
    history["runtime"] = runtimes.tolist()
    return history


def _adjust_options_to_algorithm(
    algo_options,
    lower_bounds,
    upper_bounds,
    nonlinear_constraints,
    algo_name,
    valid_kwargs,
):
    """Reduce the algo_options and check if bounds are compatible with algorithm."""

    # convert algo option keys to valid Python arguments
    algo_options = {key.replace(".", "_"): val for key, val in algo_options.items()}

    reduced = {key: val for key, val in algo_options.items() if key in valid_kwargs}

    ignored = {key: val for key, val in algo_options.items() if key not in valid_kwargs}

    if ignored:
        warnings.warn(
            "The following algo_options were ignored because they are not compatible "
            f"with {algo_name}:\n\n {ignored}"
        )

    if "lower_bounds" not in valid_kwargs and not (lower_bounds == -np.inf).all():
        raise ValueError(
            f"{algo_name} does not support lower bounds but your optimization "
            "problem has lower bounds (either because you specified them explicitly "
            "or because they were implied by other constraints)."
        )

    if "upper_bounds" not in valid_kwargs and not (upper_bounds == np.inf).all():
        raise ValueError(
            f"{algo_name} does not support upper bounds but your optimization "
            "problem has upper bounds (either because you specified them explicitly "
            "or because they were implied by other constraints)."
        )

    if "lower_bounds" in valid_kwargs:
        reduced["lower_bounds"] = lower_bounds

    if "upper_bounds" in valid_kwargs:
        reduced["upper_bounds"] = upper_bounds

    if "nonlinear_constraints" in valid_kwargs:
        reduced["nonlinear_constraints"] = nonlinear_constraints

    return reduced



"""Auxiliary functions for the pounders algorithm."""

from typing import NamedTuple, Union

import numpy as np
from scipy.linalg import qr_multiply

from estimagic.optimization.subsolvers.bntr import (
    bntr,
)
from estimagic.optimization.subsolvers.gqtpar import (
    gqtpar,
)


class ResidualModel(NamedTuple):
    intercepts: Union[np.ndarray, None] = None  # shape (n_residuals,)
    linear_terms: Union[np.ndarray, None] = None  # shape (n_residuals, n_params)
    square_terms: Union[np.ndarray, None] = (
        None  # shape (n_residuals, n_params, n_params)
    )


class MainModel(NamedTuple):
    linear_terms: Union[np.ndarray, None] = None  # shape (n_params,)
    square_terms: Union[np.ndarray, None] = None  # shape (n_params, n_params)


def create_initial_residual_model(history, accepted_index, delta):
    """Update linear and square terms of the initial residual model.

    Args:
        history (LeastSquaresHistory): Class storing history of xs, residuals, and
            critvals.
        accepted_index (int): Index in history pointing to the currently
            accepted candidate vector.
        delta (float): Trust-region radius.

    Returns:
        ResidualModel: Residual model containing the initial parameters for
             ``linear_terms`` and ``square_terms``.

    """
    center_info = {
        "x": history.get_best_x(),
        "residuals": history.get_best_residuals(),
        "radius": delta,
    }
    n_params = len(center_info["x"])
    n_residuals = center_info["residuals"].shape[0]

    indices_not_min = [i for i in range(n_params + 1) if i != accepted_index]

    x_candidate, residuals_candidate, _ = history.get_centered_entries(
        center_info=center_info,
        index=indices_not_min,
    )

    linear_terms = np.linalg.solve(x_candidate, residuals_candidate)
    square_terms = np.zeros((n_residuals, n_params, n_params))

    residual_model = ResidualModel(
        intercepts=history.get_best_residuals(),
        linear_terms=linear_terms,
        square_terms=square_terms,
    )

    return residual_model


def update_residual_model(residual_model, coefficients_to_add, delta, delta_old):
    """Update linear and square terms of the residual model.

    Args:
        residual_model (ResidualModel): Residual model with the following parameters:
            ``intercepts``, ``linear_terms``, and ``square terms``.
        coefficients_to_add (dict): Coefficients used for updating the
            parameters of the residual model.
        delta (float): Trust region radius of the current iteration.
        delta_old (float): Trust region radius of the previous iteration.

    Returns:
        ResidualModel: Residual model containing the updated parameters
            ``linear_terms`` and ``square_terms``.

    """
    linear_terms_new = (
        coefficients_to_add["linear_terms"]
        + (delta / delta_old) * residual_model.linear_terms
    )

    square_terms_new = (
        coefficients_to_add["square_terms"]
        + (delta / delta_old) ** 2 * residual_model.square_terms
    )

    residual_model_updated = residual_model._replace(
        linear_terms=linear_terms_new, square_terms=square_terms_new
    )

    return residual_model_updated


def create_main_from_residual_model(
    residual_model, multiply_square_terms_with_intercepts=True
):
    """Update linear and square terms of the main model via the residual model.

    Args:
        residual_model (ResidualModel): Residual model with the following parameters:
            ``intercepts``, ``linear_terms``, and ``square terms``.
        multiply_square_terms_with_residuals (bool): Indicator whether we
            multiply the main model's ``square terms`` with the
            intercepts of the residual model.

    Returns:
        MainModel: Main model containing the updated parameters
            ``linear_terms`` and ``square terms``.

    """
    linear_terms_main_model = residual_model.linear_terms @ residual_model.intercepts
    square_terms_main_model = (
        residual_model.linear_terms @ residual_model.linear_terms.T
    )

    if multiply_square_terms_with_intercepts is True:
        square_terms_main_model = (
            square_terms_main_model
            + residual_model.square_terms.T @ residual_model.intercepts
        )

    main_model = MainModel(
        linear_terms=linear_terms_main_model, square_terms=square_terms_main_model
    )

    return main_model


def update_main_model_with_new_accepted_x(main_model, x_candidate):
    """Use accepted candidate to update the linear terms of the residual model.

    Args:
         main_model (MainModel): Main model with the following parameters:
             ``linear_terms`` and ``square terms``.
        x_candidate (np.ndarray): Vector of centered x candidates of shape (n_params,).

    Returns:
        MainModel: Main model containing the updated ``linear_terms``.

    """
    linear_terms_new = main_model.linear_terms + main_model.square_terms @ x_candidate
    main_model_updated = main_model._replace(linear_terms=linear_terms_new)

    return main_model_updated


def update_residual_model_with_new_accepted_x(residual_model, x_candidate):
    """Use accepted candidate to update residual model.

    Args:
        residual_model (ResidualModel): Residual model containing the parameters of
            the residual model, i.e. ``intercepts``, ``linear_terms``, and
            ``square terms``.
        x_candidate (np.ndarray): Vector of centered x candidates of shape (n_params,).

    Returns:
        ResidualModel: Residual model containing the updated parameters
            `intercepts`` and ``linear_terms``.

    """
    intercepts_new = (
        residual_model.intercepts
        + x_candidate @ residual_model.linear_terms
        + 0.5 * (x_candidate.T @ residual_model.square_terms @ x_candidate)
    )

    linear_terms_new = (
        residual_model.linear_terms + (residual_model.square_terms @ x_candidate).T
    )

    residual_model_updated = residual_model._replace(
        intercepts=intercepts_new, linear_terms=linear_terms_new
    )

    return residual_model_updated


def solve_subproblem(
    x_accepted,
    main_model,
    lower_bounds,
    upper_bounds,
    delta,
    solver,
    *,
    conjugate_gradient_method,
    maxiter,
    maxiter_gradient_descent,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    gtol_abs_conjugate_gradient,
    gtol_rel_conjugate_gradient,
    k_easy,
    k_hard
):
    """Solve the quadratic subproblem.

    Args:
        x_accepted (np.ndarray): Currently accepted candidate vector of shape
            (n_params,).
        delta (float): Current trust region radius.
        main_model (MainModel): Main model with the following parameters:
             ``linear_terms`` and ``square terms``.
        lower_bounds (np.ndarray): 1d array of shape (n_params,) with lower bounds
            for the parameter vector x.
        upper_bounds (np.ndarray): 1d array of shape (n_params,) with upper bounds
            for the parameter vector x.
        delta (float) Current trust-region radius
        solver (str): Trust-region subsolver to use. Currently, two internal solvers
            are supported:
            - "bntr" (default, supports bound constraints)
            - "gqtpar" (does not support bound constraints)
        conjugate_gradient_method (str): Method for computing the conjugate gradient
            step. Available conjugate gradient methods are:
                - "cg"
                - "steihaug_toint"
                - "trsbox" (default)
        maxiter (int): Maximum number of iterations to perform when solving the
            trust-region subproblem.
        maxiter_gradient_descent (int): Maximum number of gradient descent iterations
            to perform when the trust-region subsolver "bntr" is used.
        gtol_abs (float): Convergence tolerance for the absolute gradient norm
            in the trust-region subproblem ("bntr").
        gtol_rel (float): Convergence tolerance for the relative gradient norm
            in the trust-region subproblem ("bntr").
        gtol_scaled (float): Convergence tolerance for the scaled gradient norm
            in the trust-region subproblem ("bntr").
        gtol_abs_conjugate_gradient (float): Convergence tolerance for the absolute
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("bntr").
        gtol_rel_conjugate_gradient (float): Convergence tolerance for the relative
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("bntr").
        k_easy (float): topping criterion for the "easy" case in the trust-region
            subproblem ("gqtpar").
        k_hard (float): Stopping criterion for the "hard" case in the trust-region
            subproblem ("gqtpar").

    Returns:
        (dict): Result dictionary containing the followng keys:
            - "x" (np.ndarray): The solution vector of shape (n_params,)
            - "criterion" (float): The value of the criterion functions associated
                with the solution
            - "n_iterations" (int): Number of iterations performed before termination.
            - "success" (bool): Boolean indicating whether a solution has been found
                before reaching maxiter.

    """
    x0 = np.zeros_like(x_accepted)

    # Normalize bounds. If none provided, use unit cube [-1, 1]
    if lower_bounds is not None:
        lower_bounds = (lower_bounds - x_accepted) / delta
        lower_bounds[lower_bounds < -1] = -1
    else:
        lower_bounds = -np.ones_like(x_accepted)

    if upper_bounds is not None:
        upper_bounds = (upper_bounds - x_accepted) / delta
        upper_bounds[upper_bounds > 1] = 1
    else:
        upper_bounds = np.ones_like(x_accepted)

    # Check if bounds are valid
    if np.max(lower_bounds - upper_bounds) > 1e-10:
        raise ValueError("Upper bounds < lower bounds in subproblem.")
    if np.max(lower_bounds - x0) > 1e-10:
        raise ValueError("Initial guess < lower bounds in subproblem.")
    if np.max(x0 - upper_bounds) > 1e-10:
        raise ValueError("Initial guess > upper bounds in subproblem.")

    if solver == "bntr":
        options = {
            "conjugate_gradient_method": conjugate_gradient_method,
            "maxiter": maxiter,
            "maxiter_gradient_descent": maxiter_gradient_descent,
            "gtol_abs": gtol_abs,
            "gtol_rel": gtol_rel,
            "gtol_scaled": gtol_scaled,
            "gtol_abs_conjugate_gradient": gtol_abs_conjugate_gradient,
            "gtol_rel_conjugate_gradient": gtol_rel_conjugate_gradient,
        }
        result = bntr(main_model, lower_bounds, upper_bounds, x_candidate=x0, **options)
    elif solver == "gqtpar":
        result = gqtpar(
            main_model,
            x_candidate=x0,
            k_easy=k_easy,
            k_hard=k_hard,
            maxiter=maxiter,
        )
    else:
        raise ValueError(
            "Invalid subproblem solver: {solver}. Must be one of bntr, gqtpar."
        )

    # Test bounds post-solution
    if np.max(lower_bounds - result["x"]) > 1e-5:
        raise ValueError("Subproblem solution < lower bounds.")
    if np.max(result["x"] - upper_bounds) > 1e-5:
        raise ValueError("Subproblem solution > upper bounds.")

    return result


def find_affine_points(
    history,
    x_accepted,
    model_improving_points,
    project_x_onto_null,
    delta,
    theta1,
    c,
    model_indices,
    n_modelpoints,
):
    """Find affine points.

    Args:
        history (LeastSquaresHistory): Class storing history of xs, residuals,
            and critvals.
        x_accepted (np.ndarray): Accepted solution vector of the subproblem.
            Shape (n_params,).
        model_improving_points (np.ndarray): Array of shape (n_params, n_params)
            including points to improve the main model, i.e. make the main model
            fully linear, i.e. just-identified.
            If *project_x_onto_null* is False, it is an array filled with zeros.
        project_x_onto_null (int): Indicator whether to calculate the QR
            decomposition of *model_improving_points* and multiply it
             with vector *x_projected*.
        delta (float): Delta, current trust-region radius.
        theta1 (float): Threshold for adding the current x candidate to the model.
        c (float): Threshold for acceptance of the norm of our current x candidate.
        model_indices (np.ndarray): Indices related to the candidates of x
            that are currently in the main model. Shape (2 * n_params + 1,).
        n_modelpoints (int): Current number of model points.

    Returns:
        Tuple:
        - model_improving_points (np.ndarray):  Array of shape (n_params, n_params)
            including points to improve the main model, i.e. make the main model
            fully linear, i.e. just-identified.
        - model_indices (np.ndarray): Indices related to the candidates of x
            that are currently in the main model. Shape (2 *n_params* + 1,).
        - n_modelpoints (int): Current number of model points.
        - project_x_onto_null (int): Indicator whether to calculate the QR
            decomposition of *model_improving_points* and multiply it
            with vector *x_projected*.
            Relevant for next call of *find_affine_points()*.

    """
    n_params = len(x_accepted)

    for i in range(history.get_n_fun() - 1, -1, -1):
        center_info = {"x": x_accepted, "radius": delta}
        x_candidate = history.get_centered_xs(center_info, index=i)
        candidate_norm = np.linalg.norm(x_candidate)

        x_projected = x_candidate

        if candidate_norm <= c:
            if project_x_onto_null is True:
                x_projected, _ = qr_multiply(model_improving_points, x_projected)

            proj = np.linalg.norm(x_projected[n_modelpoints:])

            # Add this index to the model
            if proj >= theta1:
                model_indices[n_modelpoints] = i
                model_improving_points[:, n_modelpoints] = x_candidate
                project_x_onto_null = True
                n_modelpoints += 1

            if n_modelpoints == n_params:
                break

    return model_improving_points, model_indices, n_modelpoints, project_x_onto_null


def add_geomtery_points_to_make_main_model_fully_linear(
    history,
    main_model,
    model_improving_points,
    model_indices,
    x_accepted,
    n_modelpoints,
    delta,
    criterion,
    lower_bounds,
    upper_bounds,
    batch_evaluator,
    n_cores,
):
    """Add points until main model is fully linear.

    Args:
        history (LeastSquaresHistory): Class storing history of xs, residuals, and
            critvals.
        main_model (MainModel): Main model with the following parameters:
             ``linear_terms`` and ``square terms``.
        model_improving_points (np.ndarray): Array of shape (n_params, n_params)
            including points to improve the main model.
        model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the main model. Shape (2 * n_params + 1,).
        x_accepted (np.ndarray): Accepted solution vector of the subproblem.
            Shape (n_params,).
        n_modelpoints (int): Current number of model points.
        delta (float): Delta, current trust-region radius.
        criterion (callable): Criterion function.
        lower_bounds (np.ndarray): Lower bounds.
            Must have same length as the initial guess of the
            parameter vector. Equal to -1 if not provided by the user.
        upper_bounds (np.ndarray): Upper bounds.
            Must have same length as the initial guess of the
            parameter vector. Equal to 1 if not provided by the user.
        batch_evaluator (str or callable): Name of a pre-implemented batch evaluator
            (currently 'joblib' and 'pathos_mp') or Callable with the same interface
            as the estimagic batch_evaluators.
        n_cores (int): Number of processes used to parallelize the function
            evaluations.

    Returns:
        Tuple:
        - history (class): Class storing history of xs, residuals, and critvals.
        - model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the main model. Shape (2 * n_params + 1,).

    """
    n_params = len(x_accepted)

    current_history = history.get_n_fun()

    x_candidate = np.zeros_like(x_accepted)
    x_candidates_list = []
    criterion_candidates_list = []

    model_improving_points, _ = qr_multiply(model_improving_points, np.eye(n_params))

    for i in range(n_modelpoints, n_params):
        change_direction = model_improving_points[:, i] @ main_model.linear_terms

        if change_direction > 0:
            model_improving_points[:, i] *= -1

        x_candidate = delta * model_improving_points[:, i] + x_accepted

        # Project into feasible region
        if lower_bounds is not None and upper_bounds is not None:
            x_candidate = np.median(
                np.stack([lower_bounds, x_candidate, upper_bounds]), axis=0
            )
        x_candidates_list.append(x_candidate)
        model_indices[i] = current_history + i - n_modelpoints

    criterion_candidates_list = batch_evaluator(
        criterion, arguments=x_candidates_list, n_cores=n_cores
    )

    history.add_entries(x_candidates_list, criterion_candidates_list)

    return history, model_indices


def evaluate_residual_model(
    centered_xs,
    centered_residuals,
    residual_model,
):
    """Compute the difference between observed and predicted model evaluations.

    We use a quadratic model of the form:

        f(x) = a + x.T @ b + 0.5 x.T @ C @ x ,

    where C is lower triangular. Note the connection of b and C to the gradient:
    f'(x) = b + (C + C.T) @ x, and the Hessian: f''(x) = C + C.T.

    Args:
        residual_model (ResidualModel): The residual model. Has entries:
            - ``intercept``: corresponds to 'a' in the above equation
            - ``linear_terms``: corresponds to 'b' in the above equation
            - ``square_terms``: corresponds to 'C' in the above equation
        centered_xs (np.ndarray): Centered x sample. Shape (n_modelpoints, n_params).
        centered_residuals (np.ndarray): Centered residuals, i.e. the observed model
            evaluations. Shape (n_maxinterp, n_residuals).

    Returns:
        np.ndarray: Observed minus predicted model evaluations,
            has shape (n_modelpoints, n_residuals).

    """
    n_residuals = centered_residuals.shape[1]
    n_modelpoints = centered_xs.shape[0]
    y_residuals = np.empty((n_modelpoints, n_residuals), dtype=np.float64)

    for j in range(n_residuals):
        x_dot_square_terms = centered_xs @ residual_model.square_terms[j, :, :]

        for i in range(n_modelpoints):
            y_residuals[i, j] = (
                centered_residuals[i, j]
                - residual_model.linear_terms[:, j] @ centered_xs[i, :]
                - 0.5 * (x_dot_square_terms[i, :] @ centered_xs[i, :])
            )

    return y_residuals


def get_feature_matrices_residual_model(
    history, x_accepted, model_indices, delta, c2, theta2, n_maxinterp
):
    """Obtain the feature matrices for fitting the residual model.

    Pounders uses underdetermined sample sets, with at most n_maxinterp
    points in the model. Hence, the fitting method is interpolation,
    where the solution represents the quadratic whose Hessian matrix is of
    minimum Frobenius norm.

    For a mathematical exposition see :cite:`Wild2008`, p. 3-5.

    Args:
        history (LeastSquaresHistory): Class storing history of xs, residuals, and
            critvals.
        x_accepted (np.ndarray): Accepted solution vector of the subproblem.
            Shape (n_params,).
        model_indices (np.ndarray): Indices of the candidates of x that are
            currently in the model. Shape (2 * n_params + 1,).
        delta (float): Delta, current trust-region radius.
        c2 (int): Threshold for acceptance of the norm of our current x candidate.
            Equal to 10 by default.
        theta2 (float): Threshold for adding the current x candidate to the model.
        n_maxinterp (int): Maximum number of interpolation points. By default,
            2 * n_params + 1 points.

    Returns:
        Tuple:
        - m_mat (np.ndarray): Polynomial feature matrix of the linear terms.
            Shape(n_params + 1, n_params + 1).
        - n_mat (np.ndarray): Polynomial feature matrix of the square terms.
            Shape(n_modelpoints, n_poly_features).
        - z_mat (np.ndarray): Basis for the null space of m_mat.
            Shape(n_modelpoints, n_modelpoints - n_params - 1).
        - n_z_mat (np.ndarray): Lower triangular matrix of xs that form
            the monomial basis. Shape(n_poly_features, n_modelpoints - n_params - 1).
        - n_modelpoints (int): Current number of model points.

    """
    n_params = len(x_accepted)
    n_poly_features = n_params * (n_params + 1) // 2

    m_mat = np.zeros((n_maxinterp, n_params + 1))
    m_mat[:, 0] = 1
    m_mat_pad = np.zeros((n_maxinterp, n_maxinterp))
    m_mat_pad[:n_maxinterp, : n_params + 1] = m_mat

    n_mat = np.zeros((n_maxinterp, n_poly_features))

    center_info = {"x": x_accepted, "radius": delta}
    for i in range(n_params + 1):
        m_mat[i, 1:] = history.get_centered_xs(center_info, index=model_indices[i])
        n_mat[i, :] = _get_monomial_basis(m_mat[i, 1:])

    point = history.get_n_fun() - 1
    n_modelpoints = n_params + 1

    while (n_modelpoints < n_maxinterp) and (point >= 0):
        reject = False

        # Reject any points already in the model
        for i in range(n_params + 1):
            if point == model_indices[i]:
                reject = True
                break

        if reject is False:
            candidate_x = history.get_centered_xs(center_info, index=point)
            candidate_norm = np.linalg.norm(candidate_x)

            if candidate_norm > c2:
                reject = True

        if reject is True:
            point -= 1
            continue

        m_mat[n_modelpoints, 1:] = history.get_centered_xs(center_info, index=point)
        n_mat[n_modelpoints, :] = _get_monomial_basis(m_mat[n_modelpoints, 1:])

        m_mat_pad = np.zeros((n_maxinterp, n_maxinterp))
        m_mat_pad[:n_maxinterp, : n_params + 1] = m_mat

        _n_z_mat, _ = qr_multiply(
            m_mat_pad[: n_modelpoints + 1, :],
            n_mat.T[:n_poly_features, : n_modelpoints + 1],
        )
        beta = np.linalg.svd(_n_z_mat.T[n_params + 1 :], compute_uv=False)

        if beta[min(n_modelpoints - n_params, n_poly_features) - 1] > theta2:
            # Accept point
            model_indices[n_modelpoints] = point
            n_z_mat = _n_z_mat

            n_modelpoints += 1

        point -= 1

    z_mat, _ = qr_multiply(
        m_mat_pad[:n_modelpoints, :],
        np.eye(n_maxinterp)[:, :n_modelpoints],
    )

    # Just-identified case
    if n_modelpoints == (n_params + 1):
        n_z_mat = np.zeros((n_maxinterp, n_poly_features))
        n_z_mat[:n_params, :n_params] = np.eye(n_params)

    return (
        m_mat[: n_params + 1, : n_params + 1],
        n_mat[:n_modelpoints],
        z_mat[:n_modelpoints, n_params + 1 : n_modelpoints],
        n_z_mat[:, n_params + 1 : n_modelpoints],
        n_modelpoints,
    )


def fit_residual_model(
    m_mat,
    n_mat,
    z_mat,
    n_z_mat,
    y_residuals,
    n_modelpoints,
):
    """Fit a linear model using the pounders fitting method.

    Pounders uses underdetermined sample sets, with at most 2 * n_params + 1
    points in the model. Hence, the fitting method is interpolation, where
    the solution represents the quadratic whose Hessian matrix is of
    minimum Frobenius norm.

    For a mathematical exposition, see :cite:`Wild2008`, p. 3-5.

    Args:
        m_mat (np.ndarray): Polynomial feature matrix of the linear terms.
            Shape(n_params + 1, n_params + 1).
        n_mat (np.ndarray): Polynomial feature matrix of the square terms.
            Shape(n_modelpoints, n_poly_features).
        z_mat (np.ndarray): Basis for the null space of m_mat.
            Shape(n_modelpoints, n_modelpoints - n_params - 1).
        n_z_mat (np.ndarray): Lower triangular matrix of xs that form
            the monomial basis. Shape(n_poly_features, n_modelpoints - n_params - 1).
        n_modelpoints (int): Current number of model points.
        y_residuals (np.ndarray): The dependent variable. Observed minus predicted
            evaluations of the residual model. Shape (n_modelpoints, n_residuals).
        n_maxinterp (int): Maximum number of interpolation points. By default,
            2 * n_params + 1 points.

    Returns:
        dict: The coefficients of the residual model.

    """
    n_params = m_mat.shape[1] - 1
    n_residuals = y_residuals.shape[1]
    n_poly_terms = n_params * (n_params + 1) // 2
    _is_just_identified = n_modelpoints == (n_params + 1)

    coeffs_linear = np.empty((n_residuals, n_params))
    coeffs_square = np.empty((n_residuals, n_params, n_params))

    if _is_just_identified:
        coeffs_first_stage = np.zeros(n_params)
        beta = np.zeros(n_poly_terms)
    else:
        n_z_mat_square = n_z_mat.T @ n_z_mat

    for k in range(n_residuals):
        if not _is_just_identified:
            z_y_vec = np.dot(z_mat.T, y_residuals[:, k])
            coeffs_first_stage = np.linalg.solve(
                np.atleast_2d(n_z_mat_square),
                np.atleast_1d(z_y_vec),
            )

            beta = np.atleast_2d(n_z_mat) @ coeffs_first_stage

        rhs = y_residuals[:, k] - n_mat @ beta

        alpha = np.linalg.solve(m_mat, rhs[: n_params + 1])
        coeffs_linear[k, :] = alpha[1 : (n_params + 1)]

        num = 0
        for i in range(n_params):
            coeffs_square[k, i, i] = beta[num]
            num += 1
            for j in range(i + 1, n_params):
                coeffs_square[k, j, i] = beta[num] / np.sqrt(2)
                coeffs_square[k, i, j] = beta[num] / np.sqrt(2)
                num += 1

    coef = {
        "linear_terms": coeffs_linear.T,
        "square_terms": coeffs_square,
    }

    return coef


def update_trustregion_radius(
    result_subproblem,
    rho,
    model_is_valid,
    delta,
    delta_min,
    delta_max,
    eta1,
    gamma0,
    gamma1,
):
    """Update the trust-region radius."""
    norm_x_sub = np.sqrt(np.sum(result_subproblem["x"] ** 2))

    if rho >= eta1 and norm_x_sub > 0.5 * delta:
        delta = min(delta * gamma1, delta_max)
    elif model_is_valid is True:
        delta = max(delta * gamma0, delta_min)

    return delta


def get_last_model_indices_and_check_for_repeated_model(
    model_indices, last_model_indices, n_modelpoints, n_last_modelpoints
):
    """Get the last model_indices and check if we have reused the same model."""
    if n_modelpoints == n_last_modelpoints:
        same_model_used = True
    else:
        same_model_used = False

    for i in range(n_modelpoints):
        if same_model_used:
            if model_indices[i] == last_model_indices[i]:
                same_model_used = True
            else:
                same_model_used = False
        last_model_indices[i] = model_indices[i]

    n_last_modelpoints = n_modelpoints

    return last_model_indices, n_last_modelpoints, same_model_used


def add_accepted_point_to_residual_model(model_indices, accepted_index, n_modelpoints):
    """Add accepted point to the residual model."""
    model_indices[1 : n_modelpoints + 1] = model_indices[:n_modelpoints]
    model_indices[0] = accepted_index

    return model_indices


def _get_monomial_basis(x):
    """Get the monomial basis (basis for quadratic functions) of x.

    Monomial basis = .5*[x(1)^2  sqrt(2)*x(1)*x(2) ... sqrt(2)*x(1)*x(n_params) ...
        ... x(2)^2 sqrt(2)*x(2)*x(3) .. x(n_params)^2]

    Args:
        x (np.ndarray): Parameter vector of shape (n_params,).

    Returns:
        np.ndarray: Monomial basis of x of shape (n_params * (n_params + 1) / 2,).

    """
    n_params = len(x)
    monomial_basis = np.zeros(int(n_params * (n_params + 1) / 2))

    j = 0
    for i in range(n_params):
        monomial_basis[j] = 0.5 * x[i] ** 2
        j += 1

        for k in range(i + 1, n_params):
            monomial_basis[j] = x[i] * x[k] / np.sqrt(2)
            j += 1

    return monomial_basis



"""Implement `simopt` optimizers.

.. note::     `simopt`'s SPSA and STRONG support box constraints, i.e. parameter bounds.
However,     for the moment they are not supported.

"""

import numpy as np

from estimagic.config import IS_SIMOPT_INSTALLED
from estimagic.decorators import mark_minimizer
from estimagic.logging.read_from_database import list_of_dicts_to_dict_of_lists
from estimagic.optimization.algo_options import (
    STOPPING_MAX_CRITERION_EVALUATIONS_GLOBAL,
)


@mark_minimizer(
    name="simopt_adam",
    primary_criterion_entry="value",
    needs_scaling=True,
    is_available=IS_SIMOPT_INSTALLED,
)
def simopt_adam(
    criterion,
    derivative,
    x,
    lower_bounds,
    upper_bounds,
    *,
    stopping_max_iterations=STOPPING_MAX_CRITERION_EVALUATIONS_GLOBAL,
    crn_across_solns=True,
    r=1,
    beta_1=0.9,
    beta_2=0.999,
    alpha=1.0,
    epsilon=10e-8,
    sensitivity=10e-7,
):
    """Minimize a scalar function using the ADAM algorithm from SimOpt.

    For details see
    :ref: `list_of_scipy_algorithms`.

    """
    solver_options = {
        "crn_across_solns": crn_across_solns,
        "r": r,
        "beta_1": beta_1,
        "beta_2": beta_2,
        "alpha": alpha,
        "epsilon": epsilon,
        "sensitivity": sensitivity,
    }

    out = _minimize_simopt(
        algorithm="ADAM",
        criterion=criterion,
        derivative=derivative,
        x=x,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        solver_options=solver_options,
        budget=stopping_max_iterations,
    )
    return out


@mark_minimizer(
    name="simopt_astrodf",
    primary_criterion_entry="value",
    needs_scaling=True,
    is_available=IS_SIMOPT_INSTALLED,
)
def simopt_astrodf(
    criterion,
    x,
    lower_bounds,
    upper_bounds,
    *,
    stopping_max_iterations=STOPPING_MAX_CRITERION_EVALUATIONS_GLOBAL,
    bounds_padding=1e-8,
    crn_across_solns=True,
    delta_max=50.0,
    eta_1=0.1,
    eta_2=0.5,
    gamma_1=2.0,
    gamma_2=0.5,
    w=0.85,
    mu=1_000.0,
    beta=10.0,
    lambda_min=None,
    simple_solve=False,
    criticality_select=True,
    criticality_threshold=0.1,
):
    """Minimize a scalar function using the ASTRODF algorithm from SimOpt.

    For details see
    :ref: `list_of_scipy_algorithms`.

    """
    solver_options = {
        "crn_across_solns": crn_across_solns,
        "delta_max": delta_max,
        "eta_1": eta_1,
        "eta_2": eta_2,
        "gamma_1": gamma_1,
        "gamma_2": gamma_2,
        "w": w,
        "mu": mu,
        "beta": beta,
        "lambda_min": 2 * len(x) + 1 if lambda_min is None else lambda_min,
        "simple_solve": simple_solve,
        "criticality_select": criticality_select,
        "criticality_threshold": criticality_threshold,
    }

    # Revert bounds shifting of ASTRODF to improve accuracy. For details see docstring.
    lower_bounds -= 0.01 - bounds_padding
    upper_bounds += 0.01 + bounds_padding

    out = _minimize_simopt(
        algorithm="ASTRODF",
        criterion=criterion,
        derivative=None,
        x=x,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        solver_options=solver_options,
        budget=stopping_max_iterations,
    )
    return out


@mark_minimizer(
    name="simopt_spsa",
    primary_criterion_entry="value",
    needs_scaling=True,
    is_available=IS_SIMOPT_INSTALLED,
)
def simopt_spsa(
    criterion,
    x,
    *,
    stopping_max_iterations=STOPPING_MAX_CRITERION_EVALUATIONS_GLOBAL,
    crn_across_solns=True,
    alpha=0.602,
    gamma=0.101,
    step=0.5,
    gavg=1,
    n_reps=2,
    n_loss=2,
    eval_pct=2 / 3,
    iter_pct=0.1,
):
    """Minimize a scalar function using the SPSA algorithm from SimOpt.

    For details see
    :ref: `list_of_scipy_algorithms`.

    """
    solver_options = {
        "crn_across_solns": crn_across_solns,
        "alpha": alpha,
        "gamma": gamma,
        "step": step,
        "gavg": gavg,
        "n_reps": n_reps,
        "n_loss": n_loss,
        "eval_pct": eval_pct,
        "iter_pct": iter_pct,
    }

    out = _minimize_simopt(
        algorithm="SPSA",
        criterion=criterion,
        derivative=None,
        x=x,
        lower_bounds=np.full_like(x, -np.inf),
        upper_bounds=np.full_like(x, np.inf),
        solver_options=solver_options,
        budget=stopping_max_iterations,
    )
    return out


@mark_minimizer(
    name="simopt_strong",
    primary_criterion_entry="value",
    needs_scaling=True,
    is_available=IS_SIMOPT_INSTALLED,
)
def simopt_strong(
    criterion,
    x,
    *,
    stopping_max_iterations=STOPPING_MAX_CRITERION_EVALUATIONS_GLOBAL,
    crn_across_solns=True,
    n0=10,
    n_r=1,
    sensitivity=10e-7,
    delta_threshold=1.2,
    delta_T=2.0,  # noqa: N803
    eta_0=0.01,
    eta_1=0.3,
    gamma_1=0.9,
    gamma_2=1.11,
    _lambda=2,
    lambda_2=1.01,
):
    """Minimize a scalar function using the STRONG algorithm from SimOpt.

    For details see
    :ref: `list_of_scipy_algorithms`.

    """
    solver_options = {
        "crn_across_solns": crn_across_solns,
        "n0": n0,
        "n_r": n_r,
        "sensitivity": sensitivity,
        "delta_threshold": delta_threshold,
        "delta_T": delta_T,
        "eta_0": eta_0,
        "eta_1": eta_1,
        "gamma_1": gamma_1,
        "gamma_2": gamma_2,
        "lambda": _lambda,
        "lambda_2": lambda_2,
    }

    out = _minimize_simopt(
        algorithm="STRONG",
        criterion=criterion,
        derivative=None,
        x=x,
        lower_bounds=np.full_like(x, -np.inf),
        upper_bounds=np.full_like(x, np.inf),
        solver_options=solver_options,
        budget=stopping_max_iterations,
    )
    return out


# ======================================================================================
# Template
# ======================================================================================


def _minimize_simopt(
    algorithm,
    criterion,
    derivative,
    x,
    lower_bounds,
    upper_bounds,
    solver_options,
    budget,
):
    problem = ProblemSpecification(
        criterion=criterion,
        derivative=derivative,
        x=x,
        lower_bounds=lower_bounds,
        upper_bounds=upper_bounds,
        budget=budget,
    )
    solver = ProblemSolver(
        solver_name=algorithm, problem=problem, solver_fixed_factors=solver_options
    )
    # overwrite method of simopt ProblemSolver class that pickles temporary results
    solver.record_experiment_results = _do_nothing.__get__(solver, ProblemSolver)
    solver.run(n_macroreps=1)

    criterion_history = criterion.keywords["history_container"]
    criterion_history = list_of_dicts_to_dict_of_lists(criterion_history)
    arg_min = np.argmin(criterion_history["criterion"])

    out = {
        "solution_x": criterion_history["params"][arg_min],
        "solution_criterion": criterion_history["criterion"][arg_min],
        "n_criterion_evaluations": len(criterion_history["criterion"]),
    }
    if derivative is not None:
        derivative_history = derivative.keywords["history_container"]
        derivative_history = list_of_dicts_to_dict_of_lists(derivative_history)
        out["solution_derivative"] = derivative_history["criterion"][arg_min]
        out["n_derivative_evaluations"] = len(derivative_history["criterion"])

    return out


def _do_nothing(self):  # noqa: ARG001
    pass


# ======================================================================================
# SimOpt Classes
# ======================================================================================


if IS_SIMOPT_INSTALLED:
    from simopt.base import Model, Problem
    from simopt.experiment_base import ProblemSolver

    class ProblemSpecification(Problem):
        def __init__(
            self,
            criterion,
            derivative,
            x,
            lower_bounds,
            upper_bounds,
            budget,
            fixed_factors=None,
        ):
            fixed_factors = {} if fixed_factors is None else fixed_factors
            self.name = "ProblemSpecification"
            self.dim = len(x)
            self.n_objectives = 1
            self.n_stochastic_constraints = 0
            self.minmax = (-1,)  # minimize objective
            self.lower_bounds = lower_bounds
            self.upper_bounds = upper_bounds
            self.gradient_available = derivative is not None
            self.model_default_factors = {}
            self.specifications = {
                "initial_solution": {"default": x},
                "budget": {"default": budget},
            }
            super().__init__(fixed_factors, model_fixed_factors={})
            self.model = CriterionWrapper(
                criterion=criterion,
                gradient=derivative,
                x=x,
            )

        def vector_to_factor_dict(self, x):
            return {"x": x}

        def factor_dict_to_vector(self, d):
            return d["x"]

        def response_dict_to_objectives(self, d):
            return (d["value"],)

    class CriterionWrapper(Model):
        def __init__(self, criterion, gradient, x):
            self.n_rngs = 0
            self.n_responses = 1
            self.specifications = {"x": {"default": x}}
            self.criterion = criterion
            self.gradient_available = gradient is not None
            self.gradient = gradient
            super().__init__(fixed_factors={})

        def replicate(self, rng_list):  # noqa: ARG002
            x = np.array(self.factors["x"])
            criterion = {"value": self.criterion(x)}
            if self.gradient_available:
                gradient = {"value": {"x": self.gradient(x)}}
            else:
                gradient = None
            return criterion, gradient



import numpy as np
import pandas as pd


def process_benchmark_results(
    problems, results, stopping_criterion, x_precision=1e-4, y_precision=1e-4
):
    """Create tidy DataFrame with all information needed for the benchmarking plots.

    Args:
        problems (dict): estimagic benchmarking problems dictionary. Keys are the
            problem names. Values contain information on the problem, including the
            solution value.
        results (dict): estimagic benchmarking results dictionary. Keys are
            tuples of the form (problem, algorithm), values are dictionaries of the
            collected information on the benchmark run, including 'criterion_history'
            and 'time_history'.
        stopping_criterion (str): one of "x_and_y", "x_or_y", "x", "y", or None.
            Determines how convergence is determined from the two precisions.
            If None, no convergence criterion is applied.
        x_precision (float): how close an algorithm must have gotten to the
            true parameter values (as percent of the Euclidean distance between start
            and solution parameters) before the criterion for clipping and convergence
            is fulfilled. Default is 1e-4.
        y_precision (float): how close an algorithm must have gotten to the
            true criterion values (as percent of the distance between start
            and solution criterion value) before the criterion for clipping and
            convergence is fulfilled. Default is 1e-4.

    Returns:
        pandas.DataFrame: tidy DataFrame with the following columns:
            - problem
            - algorithm
            - n_evaluations
            - walltime
            - criterion
            - criterion_normalized
            - monotone_criterion
            - monotone_criterion_normalized
            - parameter_distance
            - parameter_distance_normalized
            - monotone_parameter_distance
            - monotone_parameter_distance_normalized

    """
    histories = []
    infos = []

    for (problem_name, algorithm_name), result in results.items():
        history, is_converged = _process_one_result(
            problem=problems[problem_name],
            result=result,
            stopping_criterion=stopping_criterion,
            x_precision=x_precision,
            y_precision=y_precision,
        )
        history["problem"] = problem_name
        history["algorithm"] = algorithm_name
        histories.append(history)

        info = {
            "problem": problem_name,
            "algorithm": algorithm_name,
            "is_converged": is_converged,
        }
        infos.append(info)

    # breakpoint()
    histories = pd.concat(histories, ignore_index=True)
    infos = pd.DataFrame(infos).set_index(["problem", "algorithm"]).unstack()
    infos.columns = [tup[1] for tup in infos.columns]

    return histories, infos


def _process_one_result(
    problem,
    result,
    stopping_criterion,
    x_precision,
    y_precision,
):
    # input processing
    assert isinstance(x_precision, float)
    assert isinstance(y_precision, float)

    # extract information
    _params_hist = result["params_history"]
    _solution_crit = problem["solution"]["value"]
    _start_crit = problem["start_criterion"]
    _solution_x = problem["solution"].get("params")
    _start_x = problem["inputs"]["params"]
    _needed_step = np.linalg.norm(_solution_x - _start_x)
    if isinstance(_solution_x, np.ndarray) and not np.isfinite(_solution_x).all():
        _solution_x = None

    # calculate the different transformations of criterion values
    crit_hist = np.array(result["criterion_history"])
    monotone_crit_hist = np.minimum.accumulate(crit_hist)
    normalized_crit_hist = (crit_hist - _solution_crit) / (_start_crit - _solution_crit)
    normalized_monotone_crit_hist = (monotone_crit_hist - _solution_crit) / (
        _start_crit - _solution_crit
    )

    # calculate the different versions of params distance if we have a solution
    if _solution_x is not None:
        params_dist = np.linalg.norm(np.array(_params_hist - _solution_x), axis=1)
        monotone_params_dist = np.minimum.accumulate(params_dist)
        params_dist_normalized = params_dist / _needed_step
        monotone_params_dist_normalized = monotone_params_dist / _needed_step
    else:
        params_dist = np.full(len(_params_hist), np.nan)
        monotone_params_dist = np.full(len(_params_hist), np.nan)
        params_dist_normalized = np.full(len(_params_hist), np.nan)
        monotone_params_dist_normalized = np.full(len(_params_hist), np.nan)

    # put everything together in a dict
    out_dict = {
        "n_evaluations": np.arange(len(crit_hist)),
        "n_batches": result["batches_history"],
        "walltime": result["time_history"],
        "criterion": crit_hist,
        "criterion_normalized": normalized_crit_hist,
        "monotone_criterion": monotone_crit_hist,
        "monotone_criterion_normalized": normalized_monotone_crit_hist,
        "parameter_distance": params_dist,
        "monotone_parameter_distance": monotone_params_dist,
        "parameter_distance_normalized": params_dist_normalized,
        "monotone_parameter_distance_normalized": monotone_params_dist_normalized,
    }

    # calculate at which iteration the problem has been solved
    if stopping_criterion is not None:
        is_converged_x, x_idx = _check_convergence(params_dist_normalized, x_precision)
        is_converged_y, y_idx = _check_convergence(normalized_crit_hist, y_precision)

        flag_aggregators = {
            "x": lambda x, y: x,
            "y": lambda x, y: y,
            "x_and_y": lambda x, y: x and y,
            "x_or_y": lambda x, y: x or y,
        }

        is_converged = flag_aggregators[stopping_criterion](
            x=is_converged_x, y=is_converged_y
        )

        if is_converged:
            idx_aggregators = {
                "x": lambda x, y: x,
                "y": lambda x, y: y,
                "x_and_y": _aggregate_idxs_with_and,
                "x_or_y": _aggregate_idxs_with_or,
            }
            solution_idx = idx_aggregators[stopping_criterion](x=x_idx, y=y_idx)
            if solution_idx is not None:
                out_dict = {k: v[: solution_idx + 1] for k, v in out_dict.items()}

    # create a DataFrame and add metadata
    out = pd.DataFrame(out_dict)

    return out, is_converged


def _check_convergence(values, threshold):
    boo = values <= threshold
    if boo.any():
        is_converged = True
        idx = np.argmax(boo)
    else:
        is_converged = False
        idx = None
    return is_converged, idx


def _aggregate_idxs_with_and(x, y):
    if x is None or y is None:
        out = None
    else:
        out = max(x, y)
    return out


def _aggregate_idxs_with_or(x, y):
    if x is None and y is None:
        out = None
    elif x is None:
        out = y
    elif y is None:
        out = x
    else:
        out = min(x, y)
    return out



from functools import partial
from typing import NamedTuple, Callable

import numpy as np

from estimagic.parameters.space_conversion import InternalParams


def get_scale_converter(
    internal_params,
    scaling,
    scaling_options,
):
    """Get a converter between scaled and unscaled parameters.

    Args:
        internal_params (InternalParams): NamedTuple of internal and possibly
            reparametrized but not yet scaled parameter values and bounds.
        func (callable): The criterion function. Possibly used to calculate a scaling
            factor.
        scaling (bool): Whether scaling should be done.
        scaling_options (dict): User provided scaling options.

    Returns:
        ScaleConverter: NamedTuple with methods to convert between scaled and unscaled
            internal parameters and derivatives.
        InternalParams: NamedTuple with entries:
            - value (np.ndarray): Internal parameter values.
            - lower_bounds (np.ndarray): Lower bounds on the internal params.
            - upper_bounds (np.ndarray): Upper bounds on the internal params.
            - soft_lower_bounds (np.ndarray): Soft lower bounds on the internal params.
            - soft_upper_bounds (np.ndarray): Soft upper bounds on the internal params.
            - name (list): List of names of the external parameters.
            - free_mask (np.ndarray): Boolean mask representing which external parameter
              is free.

    """
    # fast path
    if not scaling:
        return _fast_path_scale_converter(), internal_params

    scaling_options = {} if scaling_options is None else scaling_options
    valid_keys = {"method", "clipping_value", "magnitude"}
    scaling_options = {k: v for k, v in scaling_options.items() if k in valid_keys}

    factor, offset = calculate_scaling_factor_and_offset(
        internal_params=internal_params, **scaling_options
    )

    _params_to_internal = partial(
        scale_to_internal,
        scaling_factor=factor,
        scaling_offset=offset,
    )

    _params_from_internal = partial(
        scale_from_internal,
        scaling_factor=factor,
        scaling_offset=offset,
    )

    def _derivative_to_internal(derivative):
        return derivative * factor

    def _derivative_from_internal(derivative):
        return derivative / factor

    converter = ScaleConverter(
        params_to_internal=_params_to_internal,
        params_from_internal=_params_from_internal,
        derivative_to_internal=_derivative_to_internal,
        derivative_from_internal=_derivative_from_internal,
    )

    if internal_params.soft_lower_bounds is not None:
        _soft_lower = converter.params_to_internal(internal_params.soft_lower_bounds)
    else:
        _soft_lower = None

    if internal_params.soft_upper_bounds is not None:
        _soft_upper = converter.params_to_internal(internal_params.soft_upper_bounds)
    else:
        _soft_upper = None

    params = InternalParams(
        values=converter.params_to_internal(internal_params.values),
        lower_bounds=converter.params_to_internal(internal_params.lower_bounds),
        upper_bounds=converter.params_to_internal(internal_params.upper_bounds),
        names=internal_params.names,
        soft_lower_bounds=_soft_lower,
        soft_upper_bounds=_soft_upper,
    )

    return converter, params


class ScaleConverter(NamedTuple):
    params_to_internal: Callable
    params_from_internal: Callable
    derivative_to_internal: Callable
    derivative_from_internal: Callable


def _fast_path_scale_converter():
    converter = ScaleConverter(
        params_to_internal=lambda x: x,
        params_from_internal=lambda x: x,
        derivative_to_internal=lambda x: x,
        derivative_from_internal=lambda x: x,
    )
    return converter


def calculate_scaling_factor_and_offset(
    internal_params,
    method="start_values",
    clipping_value=0.1,
    magnitude=1,
):
    x = internal_params.values
    lower_bounds = internal_params.lower_bounds
    upper_bounds = internal_params.upper_bounds

    if method == "start_values":
        raw_factor = np.clip(np.abs(x), clipping_value, np.inf)
        scaling_offset = None
    elif method == "bounds":
        raw_factor = upper_bounds - lower_bounds
        scaling_offset = lower_bounds

    else:
        raise ValueError(f"Invalid scaling method: {method}")

    scaling_factor = raw_factor / magnitude

    return scaling_factor, scaling_offset


def scale_to_internal(vec, scaling_factor, scaling_offset):
    """Scale a parameter vector from external scale to internal one.

    Args:
        vec (np.ndarray): Internal parameter vector with external scale.
        scaling_factor (np.ndarray or None): If None, no scaling factor is used.
        scaling_offset (np.ndarray or None): If None, no scaling offset is used.

    Returns:
        np.ndarray: vec with internal scale

    """
    if scaling_offset is not None:
        vec = vec - scaling_offset

    if scaling_factor is not None:
        vec = vec / scaling_factor

    return vec


def scale_from_internal(vec, scaling_factor, scaling_offset):
    """Scale a parameter vector from internal scale to external one.

    Args:
        vec (np.ndarray): Internal parameter vector with external scale.
        scaling_factor (np.ndarray or None): If None, no scaling factor is used.
        scaling_offset (np.ndarray or None): If None, no scaling offset is used.

    Returns:
        np.ndarray: vec with external scale

    """
    if scaling_factor is not None:
        vec = vec * scaling_factor

    if scaling_offset is not None:
        vec = vec + scaling_offset

    return vec



import pandas as pd

from estimagic.exceptions import INVALID_INFERENCE_MSG
from estimagic.inference.shared import process_pandas_arguments
from estimagic.utilities import robust_inverse


def cov_robust(jac, weights, moments_cov):
    """Calculate the cov of msm estimates with asymptotically non-efficient weights.

    Note that asymptotically non-efficient weights are typically preferrable because
    they lead to less finite sample bias.

    Args:
        jac (np.ndarray or pandas.DataFrame): Numpy array or DataFrame with the jacobian
            of simulate_moments with respect to params. The derivative needs to be taken
            at the estimated parameters. Has shape n_moments, n_params.
        weights (np.ndarray): The weighting matrix for msm estimation.
        moments_cov (np.ndarray): The covariance matrix of the empirical moments.

    Returns:
        numpy.ndarray: numpy array with covariance matrix.

    """
    _jac, _weights, _moments_cov, names = process_pandas_arguments(
        jac=jac, weights=weights, moments_cov=moments_cov
    )

    bread = robust_inverse(
        _jac.T @ _weights @ _jac,
        msg=INVALID_INFERENCE_MSG,
    )

    butter = _jac.T @ _weights @ _moments_cov @ _weights @ _jac

    cov = bread @ butter @ bread

    if names:
        cov = pd.DataFrame(cov, columns=names.get("params"), index=names.get("params"))

    return cov


def cov_optimal(jac, weights):
    """Calculate the cov of msm estimates with asymptotically efficient weights.

    Note that asymptotically efficient weights have substantial finite sample
    bias and are typically not a good choice.

    Args:
        jac (np.ndarray or pandas.DataFrame): Numpy array or DataFrame with the jacobian
            of simulate_moments with respect to params. The derivative needs to be taken
            at the estimated parameters. Has shape n_moments, n_params.
        weights (np.ndarray): The weighting matrix for msm estimation.
        moments_cov (np.ndarray): The covariance matrix of the empirical moments.

    Returns:
        numpy.ndarray: numpy array with covariance matrix.

    """
    _jac, _weights, names = process_pandas_arguments(jac=jac, weights=weights)

    cov = robust_inverse(_jac.T @ _weights @ _jac, msg=INVALID_INFERENCE_MSG)

    if names:
        cov = pd.DataFrame(cov, columns=names.get("params"), index=names.get("params"))

    return cov



import pandas as pd


def check_inputs(
    data=None, cluster_by=None, ci_method="percentile", ci_level=0.95, skipdata=False
):
    """Check validity of inputs.

    Args:
        data (pd.DataFrame): Dataset.
        cluster_by (str): Column name of variable to cluster by.
        ci_method (str): Method of choice for computing confidence intervals.
            The default is "percentile".
        ci_level (float): Confidence level for the calculation of confidence
            intervals. The default is 0.95.
        skipdata (bool): Whether to skip all checks on the data argument.

    """
    ci_method_list = ["percentile", "bc", "t", "normal", "basic"]

    if not skipdata:
        if not isinstance(data, pd.DataFrame) and not isinstance(data, pd.Series):
            raise TypeError("Data must be a pandas.DataFrame or pandas.Series.")
        elif (cluster_by is not None) and (cluster_by not in data.columns.tolist()):
            raise ValueError(
                "Input 'cluster_by' must be None or a column name of 'data'."
            )

    if ci_method not in ci_method_list:
        msg = (
            "ci_method must be 'percentile', 'bc', 't', 'basic' or 'normal', "
            f"'{ci_method}' was supplied"
        )
        raise ValueError(msg)
    if ci_level > 1 or ci_level < 0:
        raise ValueError("Input 'ci_level' must be in [0,1].")



import numpy as np
import pandas as pd
import plotly.express as px

from estimagic.benchmarking.process_benchmark_results import (
    process_benchmark_results,
)
from estimagic.config import PLOTLY_TEMPLATE


def profile_plot(
    problems,
    results,
    *,
    runtime_measure="n_evaluations",
    normalize_runtime=False,
    stopping_criterion="y",
    x_precision=1e-4,
    y_precision=1e-4,
    template=PLOTLY_TEMPLATE,
):
    """Compare optimizers over a problem set.

    This plot answers the question: What percentage of problems can each algorithm
    solve within a certain runtime budget?

    The runtime budget is plotted on the x axis and the share of problems each
    algorithm solved on the y axis.

    Thus, algorithms that are very specialized and perform well on some share of
    problems but are not able to solve more problems with a larger computational budget
    will have steep increases and then flat lines. Algorithms that are robust but slow,
    will have low shares in the beginning but reach very high.

    Note that failing to converge according to the given stopping_criterion and
    precisions is scored as needing an infinite computational budget.

    For details, see the description of performance and data profiles by
    Moré and Wild (2009).

    Args:
        problems (dict): estimagic benchmarking problems dictionary. Keys are the
            problem names. Values contain information on the problem, including the
            solution value.
        results (dict): estimagic benchmarking results dictionary. Keys are
            tuples of the form (problem, algorithm), values are dictionaries of the
            collected information on the benchmark run, including 'criterion_history'
            and 'time_history'.
        runtime_measure (str): "n_evaluations", "n_batches" or "walltime".
            This is the runtime until the desired convergence was reached by an
            algorithm. This is called performance measure by Moré and Wild (2009).
        normalize_runtime (bool): If True the runtime each algorithm needed for each
            problem is scaled by the time the fastest algorithm needed. If True, the
            resulting plot is what Moré and Wild (2009) called data profiles.
        stopping_criterion (str): one of "x_and_y", "x_or_y", "x", "y". Determines
            how convergence is determined from the two precisions.
        x_precision (float or None): how close an algorithm must have gotten to the
            true parameter values (as percent of the Euclidean distance between start
            and solution parameters) before the criterion for clipping and convergence
            is fulfilled.
        y_precision (float or None): how close an algorithm must have gotten to the
            true criterion values (as percent of the distance between start
            and solution criterion value) before the criterion for clipping and
            convergence is fulfilled.
        template (str): The template for the figure. Default is "plotly_white".

    Returns:
        plotly.Figure

    """
    if stopping_criterion is None:
        raise ValueError(
            "You must specify a stopping criterion for the performance plot. "
        )
    if runtime_measure not in ["walltime", "n_evaluations", "n_batches"]:
        raise ValueError(
            "Only 'walltime' or 'n_evaluations' are allowed as "
            f"runtime_measure. You specified {runtime_measure}."
        )

    df, converged_info = process_benchmark_results(
        problems=problems,
        results=results,
        stopping_criterion=stopping_criterion,
        x_precision=x_precision,
        y_precision=y_precision,
    )

    solution_times = create_solution_times(
        df,
        runtime_measure=runtime_measure,
        converged_info=converged_info,
    )

    if normalize_runtime:
        solution_times = solution_times.divide(solution_times.min(axis=1), axis=0)
        solution_times[~converged_info] = np.inf

    alphas = _determine_alpha_grid(solution_times)
    for_each_alpha = pd.concat(
        {alpha: solution_times <= alpha for alpha in alphas},
        names=["alpha"],
    )
    performance_profiles = for_each_alpha.groupby("alpha").mean().stack().reset_index()

    fig = px.line(performance_profiles, x="alpha", y=0, color="algorithm")

    xlabels = {
        (
            "n_evaluations",
            True,
        ): "Multiple of Minimal Number of Function Evaluations<br>"
        "Needed to Solve the Problem",
        (
            "walltime",
            True,
        ): "Multiple of Minimal Wall Time<br>Needed to Solve the Problem",
        (
            "n_batches",
            True,
        ): "Multiple of Minimal Number of Batches<br>"
        "Needed to Solve the Problem",
        ("n_evaluations", False): "Number of Function Evaluations",
        ("walltime", False): "Wall Time Needed to Solve the Problem",
        ("n_batches", False): "Number of Batches",
    }

    fig.update_layout(
        xaxis_title=xlabels[(runtime_measure, normalize_runtime)],
        yaxis_title="Share of Problems Solved",
        title=None,
        height=300,
        width=500,
        margin={"l": 10, "r": 10, "t": 30, "b": 10},
        template=template,
    )

    fig.add_hline(y=1)
    return fig


def create_solution_times(df, runtime_measure, converged_info, return_tidy=True):
    """Find the solution time for each algorithm and problem.

    Args:
        df (pandas.DataFrame): contains 'problem', 'algorithm' and 'runtime_measure'
            as columns.
        runtime_measure (str): 'walltime', 'n_batches' or 'n_evaluations'.
        converged_info (pandas.DataFrame): columns are the algorithms, indexes are the
            problems. The values are boolean and True when the algorithm arrived at
            the solution with the desired precision.
        return_tidy (bool): If True, the resulting DataFrame will be a tidy DataFrame
            with problem and algorithm as indexes and runtime_measure as column.
            If False, the resulting DataFrame will have problem, algorithm and
            runtime_measure as columns.

    Returns:
        solution_times (pandas.DataFrame): If return_tidy is True, indexes are the
            problems, columns are the algorithms. If return_tidy is False, columns are
            problem, algorithm and runtime_measure. The values are either the number
            of evaluations or the walltime each algorithm needed to achieve the
            desired precision. If the desired precision was not achieved the value is
            set to np.inf.

    """
    solution_times = df.groupby(["problem", "algorithm"])[runtime_measure].max()
    solution_times = solution_times.unstack()
    # We convert the dtype to float to support the use of np.inf
    solution_times = solution_times.astype(float).where(converged_info, other=np.inf)

    if not return_tidy:
        solution_times = solution_times.stack().reset_index()
        solution_times = solution_times.rename(
            columns={solution_times.columns[2]: runtime_measure}
        )

    return solution_times


def _determine_alpha_grid(solution_times):
    switch_points = _find_switch_points(solution_times=solution_times)

    point_to_right = switch_points[-1] * 1.05
    extended_switch_points = np.append(switch_points, point_to_right)
    mid_points = (extended_switch_points[:-1] + extended_switch_points[1:]) / 2
    alphas = sorted(np.append(extended_switch_points, mid_points))
    return alphas


def _find_switch_points(solution_times):
    """Determine the switch points of the performance profiles.

    Args:
        solution_times (pandas.DataFrame): columns are the names of the algorithms,
            the indexes are the problems. Values are performance measures.
            They can be either float, when normalize_runtime was True or int when the
            runtime_measure are not normalized function evaluations or datetime when
            the not normalized walltime is used.

    Returns:
        list: sorted switching points

    """
    switch_points = np.unique(solution_times.values)
    if pd.api.types.is_float_dtype(switch_points):
        switch_points += 1e-10
    switch_points = switch_points[np.isfinite(switch_points)]
    return switch_points



"""Functions to create, run and visualize optimization benchmarks.

TO-DO:
- Add other benchmark sets:
    - finish medium scale problems from https://arxiv.org/pdf/1710.11005.pdf, Page 34.
    - add scalar problems from https://github.com/AxelThevenot
- Add option for deterministic noise or wiggle.

"""

import numpy as np

from estimagic import batch_evaluators
from estimagic.algorithms import AVAILABLE_ALGORITHMS
from estimagic.optimization.optimize import minimize
from pybaum import tree_just_flatten
from estimagic.parameters.tree_registry import get_registry


def run_benchmark(
    problems,
    optimize_options,
    *,
    batch_evaluator="joblib",
    n_cores=1,
    error_handling="continue",
    max_criterion_evaluations=1_000,
    disable_convergence=True,
):
    """Run problems with different optimize options.

    Args:
        problems (dict): Nested dictionary with benchmark problems of the structure:
            {"name": {"inputs": {...}, "solution": {...}, "info": {...}}}
            where "inputs" are keyword arguments for ``minimize`` such as the criterion
            function and start parameters. "solution" contains the entries "params" and
            "value" and "info" might  contain information about the test problem.
        optimize_options (list or dict): Either a list of algorithms or a Nested
            dictionary that maps a name for optimizer settings
            (e.g. ``"lbfgsb_strict_criterion"``) to a dictionary of keyword arguments
            for arguments for ``minimize`` (e.g. ``{"algorithm": "scipy_lbfgsb",
            "algo_options": {"convergence.relative_criterion_tolerance": 1e-12}}``).
            Alternatively, the values can just be an algorithm which is then benchmarked
            at default settings.
        batch_evaluator (str or callable): See :ref:`batch_evaluators`.
        n_cores (int): Number of optimizations that is run in parallel. Note that in
            addition to that an optimizer might parallelize.
        error_handling (str): One of "raise", "continue".
        max_criterion_evaluations (int): Shortcut to set the maximum number of
            criterion evaluations instead of passing them in via algo options. In case
            an optimizer does not support this stopping criterion, we also use this as
            max iterations.
        disable_convergence (bool): If True, we set extremely strict convergence
            convergence criteria by default, such that most optimizers will exploit
            their full computation budget set by max_criterion_evaluations.

    Returns:
        dict: Nested Dictionary with information on the benchmark run. The outer keys
            are tuples where the first entry is the name of the problem and the second
            the name of the optimize options. The values are dicts with the entries:
            "params_history", "criterion_history", "time_history" and "solution".

    """
    if isinstance(batch_evaluator, str):
        batch_evaluator = getattr(
            batch_evaluators, f"{batch_evaluator}_batch_evaluator"
        )
    opt_options = _process_optimize_options(
        optimize_options,
        max_evals=max_criterion_evaluations,
        disable_convergence=disable_convergence,
    )

    minimize_arguments, keys = _get_optimization_arguments_and_keys(
        problems, opt_options
    )

    raw_results = batch_evaluator(
        func=minimize,
        arguments=minimize_arguments,
        n_cores=n_cores,
        error_handling=error_handling,
        unpack_symbol="**",
    )

    processing_arguments = []
    for name, raw_result in zip(keys, raw_results):
        processing_arguments.append(
            {"optimize_result": raw_result, "problem": problems[name[0]]}
        )

    results = batch_evaluator(
        func=_process_one_result,
        arguments=processing_arguments,
        n_cores=n_cores,
        error_handling="raise",
        unpack_symbol="**",
    )

    results = dict(zip(keys, results))

    return results


def _process_optimize_options(raw_options, max_evals, disable_convergence):
    if not isinstance(raw_options, dict):
        dict_options = {}
        for option in raw_options:
            if isinstance(option, str):
                dict_options[option] = option
            else:
                dict_options[option.__name__] = option
    else:
        dict_options = raw_options

    default_algo_options = {}
    if max_evals is not None:
        default_algo_options["stopping.max_criterion_evaluations"] = max_evals
        default_algo_options["stopping.max_iterations"] = max_evals
    if disable_convergence:
        default_algo_options["convergence.relative_criterion_tolerance"] = 1e-14
        default_algo_options["convergence.relative_params_tolerance"] = 1e-14
        default_algo_options["convergence.relative_gradient_tolerance"] = 1e-14

    out_options = {}
    for name, _option in dict_options.items():
        if not isinstance(_option, dict):
            option = {"algorithm": _option}
        else:
            option = _option.copy()

        algo_options = {**default_algo_options, **option.get("algo_options", {})}
        algo_options = {k.replace(".", "_"): v for k, v in algo_options.items()}
        option["algo_options"] = algo_options
        if isinstance(option.get("algo_options"), dict):
            option["algo_options"] = {**default_algo_options, **option["algo_options"]}
        else:
            option["algo_options"] = default_algo_options

        out_options[name] = option

    return out_options


def _get_optimization_arguments_and_keys(problems, opt_options):
    kwargs_list = []
    names = []

    for prob_name, problem in problems.items():
        for option_name, options in opt_options.items():
            algo = options["algorithm"]
            if isinstance(algo, str):
                if algo not in AVAILABLE_ALGORITHMS:
                    raise ValueError(f"Invalid algorithm: {algo}")
                else:
                    valid_options = AVAILABLE_ALGORITHMS[algo]._algorithm_info.arguments

            else:
                valid_options = algo._algorithm_info.arguments

            algo_options = options["algo_options"]
            algo_options = {k: v for k, v in algo_options.items() if k in valid_options}

            kwargs = {**options, **problem["inputs"]}
            kwargs["algo_options"] = algo_options
            kwargs_list.append(kwargs)
            names.append((prob_name, option_name))

    return kwargs_list, names


def _process_one_result(optimize_result, problem):
    """Process the result of one optimization run.

    Args:
        optimize_result (OptimizeResult): Result of one optimization run.
        problem (dict): Problem specification.

    Returns:
        dict: Processed result.

    """
    _registry = get_registry(extended=True)
    _criterion = problem["noise_free_criterion"]
    _start_x = problem["inputs"]["params"]
    _start_crit_value = _criterion(_start_x)
    if isinstance(_start_crit_value, np.ndarray):
        _start_crit_value = (_start_crit_value**2).sum()
    _is_noisy = problem["noisy"]
    _solution_crit = problem["solution"]["value"]

    # This will happen if the optimization raised an error
    if isinstance(optimize_result, str):
        params_history_flat = [tree_just_flatten(_start_x, registry=_registry)]
        criterion_history = [_start_crit_value]
        time_history = [np.inf]
        batches_history = [0]
    else:
        history = optimize_result.history
        params_history = history["params"]
        params_history_flat = [
            tree_just_flatten(p, registry=_registry) for p in params_history
        ]
        if _is_noisy:
            criterion_history = np.array([_criterion(p) for p in params_history])
            if criterion_history.ndim == 2:
                criterion_history = (criterion_history**2).sum(axis=1)
        else:
            criterion_history = history["criterion"]
        criterion_history = np.clip(criterion_history, _solution_crit, np.inf)
        batches_history = history["batches"]
        time_history = history["runtime"]

    return {
        "params_history": params_history_flat,
        "criterion_history": criterion_history,
        "time_history": time_history,
        "batches_history": batches_history,
        "solution": optimize_result,
    }



"""Check compatibility of pc with each other and with bounds and fixes.

See the module docstring of process_constraints for naming conventions.

"""

from functools import partial

import numpy as np
import pandas as pd


from estimagic.exceptions import InvalidConstraintError, InvalidParamsError
from estimagic.utilities import cov_params_to_matrix, sdcorr_params_to_matrix


def check_constraints_are_satisfied(flat_constraints, param_values, param_names):
    """Check that params satisfies all constraints.

    This should be called before the more specialized constraints are rewritten to
    linear constraints in order to get better error messages!

    We let the checks pass if all "values" are np.nan. This way `process_constraints`
    can be used on empty params DataFrames which is useful to construct templates for
    start parameters that can be filled out by the user.

    Args:
        pc (list): List of constraints with processed selectors.
        params (pd.DataFrame): See :ref:`params`

    Raises:
        ValueError if constraints are not satisfied.

    """
    # skip check if all parameters are NaN
    if not np.isfinite(param_values).any():
        return

    for constr in flat_constraints:
        typ = constr["type"]
        subset = param_values[constr["index"]]

        report = []

        _msg = partial(_get_message, constr, param_names)

        if typ == "covariance":
            cov = cov_params_to_matrix(subset)
            e, _ = np.linalg.eigh(cov)
            if not np.all(e > -1e-8):
                report.append(_msg())
        elif typ == "sdcorr":
            cov = sdcorr_params_to_matrix(subset)
            e, _ = np.linalg.eigh(cov)
            if not np.all(e > -1e-8):
                report.append(_msg())
        elif typ == "probability":
            if not np.isclose(subset.sum(), 1, rtol=0.01):
                explanation = "Probabilities do not sum to 1."
                report.append(_msg(explanation))
            if np.any(subset < 0):
                explanation = "There are negative Probabilities."
                report.append(_msg(explanation))
            if np.any(subset > 1):
                explanation = "There are probabilities larger than 1."
                report.append(_msg(explanation))
        elif typ == "fixed":
            if "value" in constr and not np.allclose(subset, constr["value"]):
                explanation = (
                    "Fixing parameters to different values than their start values "
                    "was allowed in earlier versions of estimagic but is "
                    "forbidden now. "
                )
                report.append(_msg(explanation))
        elif typ == "increasing":
            if np.any(np.diff(subset) < 0):
                report.append(_msg())
        elif typ == "decreasing":
            if np.any(np.diff(subset) > 0):
                report.append(_msg())
        elif typ == "linear":
            wsum = subset.dot(constr["weights"])
            if "lower_bound" in constr and wsum < constr["lower_bound"]:
                explanation = "Lower bound of linear constraint is violated."
                report.append(_msg(explanation))
            elif "upper_bound" in constr and wsum > constr["upper_bound"]:
                explanation = "Upper bound of linear constraint violated"
                report.append(_msg(explanation))
            elif "value" in constr and not np.isclose(wsum, constr["value"]):
                explanation = "Equality condition of linear constraint violated"
                report.append(_msg(explanation))
        elif typ == "equality":
            if len(set(subset.tolist())) > 1:
                report.append(_msg())

        report = "\n".join(report)
        if report != "":
            raise InvalidParamsError(f"Violated constraint at start params:\n{report}")


def _get_message(constraint, param_names, explanation=""):
    start = (
        f"A constraint of type '{constraint['type']}' is not fulfilled in params, "
        "please make sure that it holds for the starting values. The problem arose "
        "because:"
    )

    if explanation:
        explanation = f" {explanation.rstrip('. ')}. "

    names = [param_names[i] for i in constraint["index"]]

    end = (
        f"The names of the involved parameters are:\n{names}\n"
        "The relevant constraint is:\n"
        f"{constraint}."
    )

    msg = start + explanation + end
    return msg


def check_types(constraints):
    """Check that no invalid constraint types are requested.

    Args:
        constraints (list): List of constraints.

    Raises:
        TypeError if invalid constraint types are encountered

    """
    valid_types = {
        "covariance",
        "sdcorr",
        "linear",
        "probability",
        "increasing",
        "decreasing",
        "equality",
        "pairwise_equality",
        "fixed",
    }
    for constr in constraints:
        if constr["type"] not in valid_types:
            raise InvalidConstraintError(
                "Invalid constraint_type: {}".format(constr["type"]),
            )


def check_for_incompatible_overlaps(transformations, parnames):
    """Check that there are no overlaps between constraints that transform parameters.

    Since the constraints are already consolidated such that only those that transform
    a parameter are left and all equality constraints are already plugged in, this
    boils down to checking that no parameter appears more than once.

    Args:
        constr_info (dict): Dict of 1d numpy arrays with info about constraints.
        transformations (list): Processed transforming constraints.
        parnames (list): List of parameter names.

    """
    all_indices = []
    for constr in transformations:
        all_indices += constr["index"]

    msg = (
        "Transforming constraints such as 'covariance', 'sdcorr', 'probability' "
        "and 'linear' cannot overlap. This includes overlaps induced by equality "
        "constraints. This was violated for the following parameters:\n{}"
    )

    if len(set(all_indices)) < len(all_indices):
        unique, counts = np.unique(all_indices, return_counts=True)
        invalid_indices = unique[counts >= 2]
        invalid_names = [parnames[i] for i in invalid_indices]

        raise InvalidConstraintError(msg.format(invalid_names))


def check_fixes_and_bounds(constr_info, transformations, parnames):
    """Check fixes.

    Warn the user if he fixes a parameter to a value even though that parameter has
    a different non-nan value in params

    check that fixes are compatible with other constraints.

    Args:
        constr_info (dict): Dict of 1d numpy arrays with info about constraints.
        transformations (list): Processed transforming constraints.
        parnames (list): List of parameter names.

    """
    constr_info = constr_info.copy()
    constr_info["index"] = parnames

    prob_msg = (
        "{} constraints are incompatible with fixes or bounds. "
        "This is violated for:\n{}"
    )

    cov_msg = (
        "{} constraints are incompatible with fixes or bounds except for the first "
        "parameter. This is violated for:\n{}"
    )

    for constr in transformations:
        if constr["type"] in ["covariance", "sdcorr"]:
            subset = _iloc(dictionary=constr_info, positions=constr["index"][1:])
            if subset["is_fixed_to_value"].any():
                problematic = subset["index"][subset["is_fixed_to_value"]]
                raise InvalidConstraintError(
                    cov_msg.format(constr["type"], problematic)
                )
            finite_bounds = np.isfinite(subset["lower_bounds"]) | np.isfinite(
                subset["upper_bounds"]
            )
            if finite_bounds.any():
                problematic = subset["index"][finite_bounds]
                raise InvalidConstraintError(
                    prob_msg.format(constr["type"], problematic)
                )
        elif constr["type"] == "probability":
            subset = _iloc(dictionary=constr_info, positions=constr["index"])
            if subset["is_fixed_to_value"].any():
                problematic = subset["index"][subset["is_fixed_to_value"]]
                raise InvalidConstraintError(
                    prob_msg.format(constr["type"], problematic)
                )
            finite_bounds = np.isfinite(subset["lower_bounds"]) | np.isfinite(
                subset["upper_bounds"]
            )
            if finite_bounds.any():
                problematic = subset["index"][finite_bounds]
                raise InvalidConstraintError(
                    prob_msg.format(constr["type"], problematic)
                )

    is_invalid = constr_info["lower_bounds"] >= constr_info["upper_bounds"]
    if is_invalid.any():
        info = pd.DataFrame(
            {
                "names": parnames[is_invalid],
                "lower_bounds": constr_info["lower_bounds"][is_invalid],
                "upper_bounds": constr_info["upper_bounds"][is_invalid],
            }
        )

        msg = (
            "lower_bound must be strictly smaller than upper_bound. "
            f"This is violated for:\n{info}"
        )

        raise InvalidConstraintError(msg)


def _iloc(dictionary, positions):
    """Substitute function for DataFrame.iloc. that works for a dictionary of arrays.

    It creates a subset of the input dictionary based on the
    index values in the info list, and returns this subset as
    a dictionary with numpy arrays.

    Args:
        dictionary (dict): Dictionary of arrays.
        position (list): List, slice or array of indices.

    """
    subset = {}
    for key, value in dictionary.items():
        if isinstance(value, list) and not isinstance(positions, slice):
            subset[key] = [value[i] for i in positions]
        else:
            subset[key] = value[positions]

    return subset



import io
import warnings

import cloudpickle
import pandas as pd
import sqlalchemy as sql

from estimagic.exceptions import get_traceback


class DataBase:
    """Class containing everything to work with a logging database.

    Importantly, the class is pickle-serializable which is important to share it across
    multiple processes. Upon unpickling, it will automatically re-create an engine to
    connect to the database.

    """

    def __init__(self, metadata, path, fast_logging, engine=None):
        self.metadata = metadata
        self.path = path
        self.fast_logging = fast_logging
        if engine is None:
            self.engine = _create_engine(path, fast_logging)
        else:
            self.engine = engine

    def __reduce__(self):
        return (DataBase, (self.metadata, self.path, self.fast_logging))


def load_database(path_or_database, fast_logging=False):
    """Load or create a database from a path and configure it for our needs.

    This is the only acceptable way of loading or creating a database in estimagic!

    Args:
        path_or_database (str or pathlib.Path): Path to the database or DataBase.
        fast_logging (bool): If True, use unsafe optimizations to speed up the logging.
            If False, only use ultra safe optimizations.

    Returns:
        database (Database): Object containing everything to work with the
            database.

    """
    if isinstance(path_or_database, DataBase):
        out = path_or_database
    else:
        engine = _create_engine(path_or_database, fast_logging)
        metadata = sql.MetaData()
        _configure_reflect()
        metadata.reflect(engine)

        out = DataBase(
            metadata=metadata,
            path=path_or_database,
            fast_logging=fast_logging,
            engine=engine,
        )
    return out


def _create_engine(path, fast_logging):
    engine = sql.create_engine(f"sqlite:///{path}")
    _configure_engine(engine, fast_logging)
    return engine


def _configure_engine(engine, fast_logging):
    """Configure the sqlite engine.

    The two functions that configure the emission of the begin statement are taken from
    the sqlalchemy documentation the documentation:
    https://tinyurl.com/u9xea5z
    and are
    the recommended way of working around a bug in the pysqlite driver.

    The other function speeds up the write process. If fast_logging is False, it does so
    using only completely safe optimizations. Of fast_logging is True, it also uses
    unsafe optimizations.

    """

    @sql.event.listens_for(engine, "connect")
    def do_connect(dbapi_connection, connection_record):  # noqa: ARG001
        # disable pysqlite's emitting of the BEGIN statement entirely.
        # also stops it from emitting COMMIT before absolutely necessary.
        dbapi_connection.isolation_level = None

    @sql.event.listens_for(engine, "begin")
    def do_begin(conn):
        # emit our own BEGIN
        conn.exec_driver_sql("BEGIN DEFERRED")

    @sql.event.listens_for(engine, "connect")
    def set_sqlite_pragma(dbapi_connection, connection_record):  # noqa: ARG001
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA journal_mode = WAL")
        if fast_logging:
            cursor.execute("PRAGMA synchronous = OFF")
        else:
            cursor.execute("PRAGMA synchronous = NORMAL")
        cursor.close()


def _configure_reflect():
    """Mark all BLOB dtypes as PickleType with our custom pickle reader.

    Code ist taken from the documentation: https://tinyurl.com/y7q287jr

    """

    @sql.event.listens_for(sql.Table, "column_reflect")
    def _setup_pickletype(inspector, table, column_info):  # noqa: ARG001
        if isinstance(column_info["type"], sql.BLOB):
            column_info["type"] = sql.PickleType(pickler=RobustPickler)


class RobustPickler:
    @staticmethod
    def loads(
        data,
        fix_imports=True,  # noqa: ARG004
        encoding="ASCII",  # noqa: ARG004
        errors="strict",  # noqa: ARG004
        buffers=None,  # noqa: ARG004
    ):
        """Robust pickle loading.

        We first try to unpickle the object with pd.read_pickle. This makes no
        difference for non-pandas objects but makes the de-serialization
        of pandas objects more robust across pandas versions. If that fails, we use
        cloudpickle. If that fails, we return None but do not raise an error.

        See: https://github.com/pandas-dev/pandas/issues/16474

        """
        try:
            res = pd.read_pickle(io.BytesIO(data), compression=None)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception:
            try:
                res = cloudpickle.loads(data)
            except (KeyboardInterrupt, SystemExit):
                raise
            except Exception:
                res = None
                tb = get_traceback()
                warnings.warn(
                    f"Unable to read PickleType column from database:\n{tb}\n "
                    "The entry was replaced by None."
                )

        return res

    @staticmethod
    def dumps(
        obj, protocol=None, *, fix_imports=True, buffer_callback=None  # noqa: ARG004
    ):
        return cloudpickle.dumps(obj, protocol=protocol)



"""This module contains various decorators.

There are two kinds of decorators defined in this module which consists of either two or
three nested functions. The former are decorators without and the latter with arguments.

For more information on decorators, see this `guide
`_ on https://realpython.com










which
provides a comprehensive overview.

.. _guide:

https://realpython.com/primer-on-python-decorators/

"""

import functools
import inspect
import warnings
from typing import NamedTuple

from estimagic.exceptions import get_traceback


def catch(
    func=None,
    *,
    exception=Exception,
    exclude=(KeyboardInterrupt, SystemExit),
    onerror=None,
    default=None,
    warn=True,
    reraise=False,
):
    """Catch and handle exceptions.

    This decorator can be used with and without additional arguments.

    Args:
        exception (Exception or tuple): One or several exceptions that
            are caught and handled. By default all Exceptions are
            caught and handled.
        exclude (Exception or tuple): One or several exceptionts that
            are not caught. By default those are KeyboardInterrupt and
            SystemExit.
        onerror (None or Callable): Callable that takes an Exception
            as only argument. This is called when an exception occurs.
        default: Value that is returned when as the output of func when
            an exception occurs. Can be one of the following:
            - a constant
            - "__traceback__", in this case a string with a traceback is returned.
            - callable with the same signature as func.
        warn (bool): If True, the exception is converted to a warning.
        reraise (bool): If True, the exception is raised after handling it.

    """

    def decorator_catch(func):
        @functools.wraps(func)
        def wrapper_catch(*args, **kwargs):
            try:
                res = func(*args, **kwargs)
            except exclude:
                raise
            except exception as e:
                if onerror is not None:
                    onerror(e)

                if reraise:
                    raise e

                tb = get_traceback()

                if warn:
                    msg = f"The following exception was caught:\n\n{tb}"
                    warnings.warn(msg)

                if default == "__traceback__":
                    res = tb
                elif callable(default):
                    res = default(*args, **kwargs)
                else:
                    res = default
            return res

        return wrapper_catch

    if callable(func):
        return decorator_catch(func)
    else:
        return decorator_catch


def unpack(func=None, symbol=None):
    def decorator_unpack(func):
        if symbol is None:

            @functools.wraps(func)
            def wrapper_unpack(arg):
                return func(arg)

        elif symbol == "*":

            @functools.wraps(func)
            def wrapper_unpack(arg):
                return func(*arg)

        elif symbol == "**":

            @functools.wraps(func)
            def wrapper_unpack(arg):
                return func(**arg)

        return wrapper_unpack

    if callable(func):
        return decorator_unpack(func)
    else:
        return decorator_unpack


def switch_sign(func):
    """Switch sign of all outputs of a function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        unswitched = func(*args, **kwargs)
        if isinstance(unswitched, dict):
            switched = {key: -val for key, val in unswitched.items()}
        elif isinstance(unswitched, (tuple, list)):
            switched = []
            for entry in unswitched:
                if isinstance(entry, dict):
                    switched.append({key: -val for key, val in entry.items()})
                else:
                    switched.append(-entry)
            if isinstance(unswitched, tuple):
                switched = tuple(switched)
        else:
            switched = -unswitched
        return switched

    return wrapper


class AlgoInfo(NamedTuple):
    primary_criterion_entry: str
    name: str
    parallelizes: bool
    needs_scaling: bool
    is_available: bool
    arguments: list
    is_global: bool = False
    disable_history: bool = False


def mark_minimizer(
    func=None,
    *,
    primary_criterion_entry="value",
    name=None,
    needs_scaling=False,
    is_available=True,
    is_global=False,
    disable_history=False,
):
    """Decorator to mark a function as internal estimagic minimizer and add information.

    Args:
        func (callable): The function to be decorated
        primary_criterion_entry (str): One of "value", "contributions",
            "root_contributions" or "dict". Default: "value". This decides
            which part of the output of the user provided criterion function
            is needed by the internal optimizer.
        name (str): The name of the internal algorithm.
        parallelizes (bool): Must be True if an algorithm evaluates the criterion,
            derivative or criterion_and_derivative in parallel.
        needs_scaling (bool): Must be True if the algorithm is not reasonable
            independent of the scaling of the parameters.
        is_available (bool): Whether the algorithm is available. This is needed for
            algorithms that require optional dependencies.
        disable_history (bool): Whether the automatic history collection should be
            disabled, for example, because the algorithm does its own history
            collection.

    """
    if name is None:
        raise TypeError(
            "mark_minimizer() missing 1 required keyword-only argument: 'name'"
        )
    elif not isinstance(name, str):
        raise TypeError("name must be a string.")

    valid_entries = ["value", "contributions", "root_contributions"]
    if primary_criterion_entry not in valid_entries:
        raise ValueError(
            f"primary_criterion_entry must be one of {valid_entries} not "
            f"{primary_criterion_entry}."
        )

    if not isinstance(needs_scaling, bool):
        raise TypeError("needs_scaling must be a bool.")

    if not isinstance(is_available, bool):
        raise TypeError("is_available must be a bool.")

    if not isinstance(disable_history, bool):
        raise TypeError("disable_history must be a bool.")

    def decorator_mark_minimizer(func):
        arguments = list(inspect.signature(func).parameters)

        if isinstance(func, functools.partial):
            partialed_in = set(func.keywords)
            arguments = [a for a in arguments if a not in partialed_in]

        parallelizes = "n_cores" in arguments

        algo_info = AlgoInfo(
            primary_criterion_entry=primary_criterion_entry,
            name=name,
            parallelizes=parallelizes,
            needs_scaling=needs_scaling,
            is_available=is_available,
            arguments=arguments,
            is_global=is_global,
            disable_history=disable_history,
        )

        @functools.wraps(func)
        def wrapper_mark_minimizer(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper_mark_minimizer._algorithm_info = algo_info

        return wrapper_mark_minimizer

    if callable(func):
        return decorator_mark_minimizer(func)

    else:
        return decorator_mark_minimizer



"""Auxiliary functions for the quadratic BNTR trust-region subsolver."""

from functools import reduce
from typing import NamedTuple, Union

import numpy as np
from estimagic.optimization.subsolvers._conjugate_gradient import (
    minimize_trust_cg,
)
from estimagic.optimization.subsolvers._steihaug_toint import (
    minimize_trust_stcg,
)
from estimagic.optimization.subsolvers._trsbox import minimize_trust_trsbox

EPSILON = np.finfo(float).eps ** (2 / 3)


class ActiveBounds(NamedTuple):
    lower: Union[np.ndarray, None] = None
    upper: Union[np.ndarray, None] = None
    fixed: Union[np.ndarray, None] = None
    active: Union[np.ndarray, None] = None
    inactive: Union[np.ndarray, None] = None


def bntr(
    model,
    lower_bounds,
    upper_bounds,
    x_candidate,
    *,
    conjugate_gradient_method,
    maxiter,
    maxiter_gradient_descent,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
    gtol_abs_conjugate_gradient,
    gtol_rel_conjugate_gradient,
):
    """Minimize a bounded trust-region subproblem via Newton Conjugate Gradient method.

    The BNTR (Bounded Newton Trust Rregion) algorithm uses an active-set approach
    to solve the symmetric system of equations:

        hessian @ x = - gradient

    only for the inactive parameters of x that lie within the bounds. The active-set
    estimation employed here is based on Bertsekas (:cite:`Bertsekas1982`).

    In the main loop, BNTR globalizes the Newton step using a trust-region method
    based on the predicted versus actual reduction in the criterion function.
    The trust-region radius is increased only if the accepted step is at the
    trust-region boundary.


    Args:
        model (NamedTuple): NamedTuple containing the parameters of the
            main model, i.e.:
            - ``linear_terms`` (np.ndarray): 1d array of shape (n,)
            - ``square_terms`` (np.ndarray): 2d array of shape (n,n).
        lower_bounds (np.ndarray): 1d array of shape (n,) with lower bounds
            for the parameter vector x.
        upper_bounds (np.ndarray): 1d array of shape (n,) with upper bounds
            for the parameter vector x.
        x_candidate (np.ndarray): Initial guess for the solution of the subproblem.
        conjugate_gradient_method (str): Method for computing the conjugate gradient
            step. Available conjugate gradient methods are:
                - "cg"
                - "steihaug_toint"
                - "trsbox" (default)
        maxiter (int): Maximum number of iterations. If reached, terminate.
        maxiter_gradient_descent (int): Maximum number of steepest descent iterations
            to perform when the trust-region subsolver BNTR is used.
        gtol_abs (float): Convergence tolerance for the absolute gradient norm.
        gtol_rel (float): Convergence tolerance for the relative gradient norm.
        gtol_scaled (float): Convergence tolerance for the scaled gradient norm.
        gtol_abs_conjugate_gradient (float): Convergence tolerance for the absolute
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("BNTR").
        gtol_rel_conjugate_gradient (float): Convergence tolerance for the relative
            gradient norm in the conjugate gradient step of the trust-region
            subproblem ("BNTR").

    Returns:
        (dict): Result dictionary containing the following keys:
            - ``x`` (np.ndarray): Solution vector of the subproblem of shape (n,)
            - ``criterion`` (float): Minimum function value associated with the
                solution.
            - ``n_iterations`` (int): Number of iterations the algorithm ran before
                termination.
            - ``success`` (bool): Boolean indicating whether a solution has been found
                before reaching maxiter.

    """
    options_update_radius = {
        "eta1": 1.0e-4,
        "eta2": 0.25,
        "eta3": 0.50,
        "eta4": 0.90,
        "alpha1": 0.25,
        "alpha2": 0.50,
        "alpha3": 1.00,
        "alpha4": 2.00,
        "alpha5": 4.00,
        "min_radius": 1e-10,
        "max_radius": 1e10,
        "default_radius": 100.00,
    }

    (
        x_candidate,
        f_candidate,
        gradient_unprojected,
        hessian_bounds_inactive,
        trustregion_radius,
        active_bounds_info,
        converged,
        convergence_reason,
    ) = _take_preliminary_gradient_descent_step_and_check_for_solution(
        x_candidate,
        model,
        lower_bounds,
        upper_bounds,
        maxiter_gradient_descent,
        gtol_abs,
        gtol_rel,
        gtol_scaled,
    )

    for niter in range(maxiter + 1):
        if converged:
            break

        x_old = x_candidate
        f_old = f_candidate
        accept_step = False

        while not accept_step and not converged:
            gradient_bounds_inactive = gradient_unprojected[active_bounds_info.inactive]
            hessian_bounds_inactive = _find_hessian_submatrix_where_bounds_inactive(
                model, active_bounds_info
            )
            (
                conjugate_gradient_step,
                conjugate_gradient_step_inactive_bounds,
                cg_step_norm,
            ) = _compute_conjugate_gradient_step(
                x_candidate,
                gradient_bounds_inactive,
                hessian_bounds_inactive,
                lower_bounds,
                upper_bounds,
                active_bounds_info,
                trustregion_radius,
                conjugate_gradient_method=conjugate_gradient_method,
                gtol_abs_conjugate_gradient=gtol_abs_conjugate_gradient,
                gtol_rel_conjugate_gradient=gtol_rel_conjugate_gradient,
                options_update_radius=options_update_radius,
            )

            x_unbounded = x_candidate + conjugate_gradient_step
            x_candidate = _apply_bounds_to_x_candidate(
                x_unbounded, lower_bounds, upper_bounds
            )

            predicted_reduction = (
                _compute_predicted_reduction_from_conjugate_gradient_step(
                    conjugate_gradient_step,
                    conjugate_gradient_step_inactive_bounds,
                    gradient_unprojected,
                    gradient_bounds_inactive,
                    hessian_bounds_inactive,
                    active_bounds_info,
                )
            )

            f_candidate = _evaluate_model_criterion(
                x_candidate, model.linear_terms, model.square_terms
            )
            actual_reduction = f_old - f_candidate

            trustregion_radius_old = trustregion_radius
            (
                trustregion_radius,
                accept_step,
            ) = _update_trustregion_radius_conjugate_gradient(
                f_candidate,
                predicted_reduction,
                actual_reduction,
                cg_step_norm,
                trustregion_radius,
                options_update_radius,
            )

            if accept_step:
                gradient_unprojected = (
                    model.linear_terms + model.square_terms @ x_candidate
                )

                active_bounds_info = _get_information_on_active_bounds(
                    x_candidate,
                    gradient_unprojected,
                    lower_bounds,
                    upper_bounds,
                )
            else:
                x_candidate = x_old
                f_candidate = f_old

                if trustregion_radius == trustregion_radius_old:
                    converged = True
                    break

            converged, convergence_reason = _check_for_convergence(
                x_candidate,
                f_candidate,
                gradient_unprojected,
                model,
                lower_bounds,
                upper_bounds,
                converged,
                convergence_reason,
                niter,
                maxiter=maxiter,
                gtol_abs=gtol_abs,
                gtol_rel=gtol_rel,
                gtol_scaled=gtol_scaled,
            )

    result = {
        "x": x_candidate,
        "criterion": f_candidate,
        "n_iterations": niter,
        "success": converged,
        "message": convergence_reason,
    }

    return result


def _take_preliminary_gradient_descent_step_and_check_for_solution(
    x_candidate,
    model,
    lower_bounds,
    upper_bounds,
    maxiter_gradient_descent,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
):
    """Take a preliminary gradient descent step and check if we found a solution."""
    options_update_radius = {
        "mu1": 0.35,
        "mu2": 0.50,
        "gamma1": 0.0625,
        "gamma2": 0.5,
        "gamma3": 2.0,
        "gamma4": 5.0,
        "theta": 0.25,
        "min_radius": 1e-10,
        "max_radius": 1e10,
        "default_radius": 100.0,
    }

    converged = False
    convergence_reason = "Continue iterating."

    criterion_candidate = _evaluate_model_criterion(
        x_candidate, model.linear_terms, model.square_terms
    )

    active_bounds_info = _get_information_on_active_bounds(
        x_candidate,
        model.linear_terms,
        lower_bounds,
        upper_bounds,
    )

    gradient_unprojected = model.linear_terms + model.square_terms @ x_candidate
    gradient_projected = _project_gradient_onto_feasible_set(
        gradient_unprojected, active_bounds_info
    )

    converged, convergence_reason = _check_for_convergence(
        x_candidate,
        criterion_candidate,
        gradient_unprojected,
        model,
        lower_bounds,
        upper_bounds,
        converged,
        convergence_reason,
        niter=None,
        maxiter=None,
        gtol_abs=gtol_abs,
        gtol_rel=gtol_rel,
        gtol_scaled=gtol_scaled,
    )

    if converged:
        hessian_inactive = model.square_terms
        trustregion_radius = options_update_radius["default_radius"]
    else:
        hessian_inactive = _find_hessian_submatrix_where_bounds_inactive(
            model, active_bounds_info
        )

        (
            x_candidate_gradient_descent,
            f_min_gradient_descent,
            step_size_gradient_descent,
            trustregion_radius,
            radius_lower_bound,
        ) = _perform_gradient_descent_step(
            x_candidate,
            criterion_candidate,
            gradient_projected,
            hessian_inactive,
            model,
            lower_bounds,
            upper_bounds,
            active_bounds_info,
            maxiter_gradient_descent,
            options_update_radius,
        )

        if f_min_gradient_descent < criterion_candidate:
            criterion_candidate = f_min_gradient_descent

            x_unbounded = (
                x_candidate_gradient_descent
                - step_size_gradient_descent * gradient_projected
            )
            x_candidate = _apply_bounds_to_x_candidate(
                x_unbounded, lower_bounds, upper_bounds
            )

            gradient_unprojected = model.linear_terms + model.square_terms @ x_candidate
            active_bounds_info = _get_information_on_active_bounds(
                x_candidate,
                gradient_unprojected,
                lower_bounds,
                upper_bounds,
            )

            gradient_projected = _project_gradient_onto_feasible_set(
                gradient_unprojected, active_bounds_info
            )
            hessian_inactive = _find_hessian_submatrix_where_bounds_inactive(
                model, active_bounds_info
            )

            converged, convergence_reason = _check_for_convergence(
                x_candidate,
                criterion_candidate,
                gradient_projected,
                model,
                lower_bounds,
                upper_bounds,
                converged,
                convergence_reason,
                niter=None,
                maxiter=None,
                gtol_abs=gtol_abs,
                gtol_rel=gtol_rel,
                gtol_scaled=gtol_scaled,
            )

        if not converged:
            trustregion_radius = np.clip(
                max(trustregion_radius, radius_lower_bound),
                options_update_radius["min_radius"],
                options_update_radius["max_radius"],
            )

    return (
        x_candidate,
        criterion_candidate,
        gradient_unprojected,
        hessian_inactive,
        trustregion_radius,
        active_bounds_info,
        converged,
        convergence_reason,
    )


def _compute_conjugate_gradient_step(
    x_candidate,
    gradient_inactive,
    hessian_inactive,
    lower_bounds,
    upper_bounds,
    active_bounds_info,
    trustregion_radius,
    *,
    conjugate_gradient_method,
    gtol_abs_conjugate_gradient,
    gtol_rel_conjugate_gradient,
    options_update_radius,
):
    """Compute the bounded Conjugate Gradient trust-region step."""
    conjugate_gradient_step = np.zeros_like(x_candidate)

    if active_bounds_info.inactive.size == 0:
        # Save some computation and return an adjusted zero step
        step_inactive = _apply_bounds_to_x_candidate(
            x_candidate, lower_bounds, upper_bounds
        )
        step_norm = np.linalg.norm(step_inactive)

        conjugate_gradient_step = _apply_bounds_to_conjugate_gradient_step(
            step_inactive,
            x_candidate,
            lower_bounds,
            upper_bounds,
            active_bounds_info,
        )

    else:
        if conjugate_gradient_method == "cg":
            step_inactive = minimize_trust_cg(
                gradient_inactive,
                hessian_inactive,
                trustregion_radius,
                gtol_abs=gtol_abs_conjugate_gradient,
                gtol_rel=gtol_rel_conjugate_gradient,
            )
            step_norm = np.linalg.norm(step_inactive)
        elif conjugate_gradient_method == "steihaug_toint":
            step_inactive = minimize_trust_stcg(
                gradient_inactive,
                hessian_inactive,
                trustregion_radius,
            )
            step_norm = np.linalg.norm(step_inactive)
        elif conjugate_gradient_method == "trsbox":
            step_inactive = minimize_trust_trsbox(
                gradient_inactive,
                hessian_inactive,
                trustregion_radius,
                lower_bounds=lower_bounds[active_bounds_info.inactive],
                upper_bounds=upper_bounds[active_bounds_info.inactive],
            )
            step_norm = np.linalg.norm(step_inactive)
        else:
            raise ValueError(
                "Invalid method: {conjugate_gradient_method}. "
                "Must be one of cg, steihaug_toint, trsbox."
            )

        if trustregion_radius == 0:
            if step_norm > 0:
                # Accept
                trustregion_radius = np.clip(
                    step_norm,
                    options_update_radius["min_radius"],
                    options_update_radius["max_radius"],
                )

            else:
                # Re-solve
                trustregion_radius = np.clip(
                    options_update_radius["default_radius"],
                    options_update_radius["min_radius"],
                    options_update_radius["max_radius"],
                )

                if conjugate_gradient_method == "cg":
                    step_inactive = minimize_trust_cg(
                        gradient_inactive,
                        hessian_inactive,
                        trustregion_radius,
                        gtol_abs=gtol_abs_conjugate_gradient,
                        gtol_rel=gtol_rel_conjugate_gradient,
                    )
                    step_norm = np.linalg.norm(step_inactive)
                elif conjugate_gradient_method == "steihaug_toint":
                    step_inactive = minimize_trust_stcg(
                        gradient_inactive,
                        hessian_inactive,
                        trustregion_radius,
                    )
                    step_norm = np.linalg.norm(step_inactive)
                elif conjugate_gradient_method == "trsbox":
                    step_inactive = minimize_trust_trsbox(
                        gradient_inactive,
                        hessian_inactive,
                        trustregion_radius,
                        lower_bounds=lower_bounds[active_bounds_info.inactive],
                        upper_bounds=upper_bounds[active_bounds_info.inactive],
                    )
                    step_norm = np.linalg.norm(step_inactive)

                if step_norm == 0:
                    raise ValueError("Initial direction is zero.")

        conjugate_gradient_step = _apply_bounds_to_conjugate_gradient_step(
            step_inactive,
            x_candidate,
            lower_bounds,
            upper_bounds,
            active_bounds_info,
        )

    return (
        conjugate_gradient_step,
        step_inactive,
        step_norm,
    )


def _compute_predicted_reduction_from_conjugate_gradient_step(
    conjugate_gradient_step,
    conjugate_gradient_step_inactive,
    gradient_unprojected,
    gradient_inactive,
    hessian_inactive,
    active_bounds_info,
):
    """Compute predicted reduction induced by the Conjugate Gradient step."""
    if active_bounds_info.active.size > 0:
        # Projection changed the step, so we have to recompute the step
        # and the predicted reduction. Leave the rust radius unchanged.
        cg_step_recomp = conjugate_gradient_step[active_bounds_info.inactive]
        gradient_inactive_recomp = gradient_unprojected[active_bounds_info.inactive]

        predicted_reduction = _evaluate_model_criterion(
            cg_step_recomp, gradient_inactive_recomp, hessian_inactive
        )
    else:
        # Step did not change, so we can just recover the
        # pre-computed prediction
        predicted_reduction = _evaluate_model_criterion(
            conjugate_gradient_step_inactive,
            gradient_inactive,
            hessian_inactive,
        )

    return -predicted_reduction


def _perform_gradient_descent_step(
    x_candidate,
    f_candidate_initial,
    gradient_projected,
    hessian_inactive,
    model,
    lower_bounds,
    upper_bounds,
    active_bounds_info,
    maxiter_steepest_descent,
    options_update_radius,
):
    """Perform gradient descent step and update trust-region radius."""
    f_min = f_candidate_initial
    gradient_norm = np.linalg.norm(gradient_projected)

    trustregion_radius = options_update_radius["default_radius"]
    radius_lower_bound = 0
    step_size_accepted = 0

    for _ in range(maxiter_steepest_descent):
        x_old = x_candidate

        step_size_candidate = trustregion_radius / gradient_norm
        x_candidate = x_old - step_size_candidate * gradient_projected

        x_candidate = _apply_bounds_to_x_candidate(
            x_candidate, lower_bounds, upper_bounds
        )
        f_candidate = _evaluate_model_criterion(
            x_candidate, model.linear_terms, model.square_terms
        )

        x_diff = x_candidate - x_old

        if f_candidate < f_min:
            f_min = f_candidate
            step_size_accepted = step_size_candidate

        x_inactive = x_diff[active_bounds_info.inactive]
        square_terms = x_inactive.T @ hessian_inactive @ x_inactive

        predicted_reduction = trustregion_radius * (
            gradient_norm - 0.5 * trustregion_radius * square_terms / (gradient_norm**2)
        )
        actual_reduction = f_candidate_initial - f_candidate

        (
            trustregion_radius,
            radius_lower_bound,
        ) = _update_trustregion_radius_and_gradient_descent(
            trustregion_radius,
            radius_lower_bound,
            predicted_reduction,
            actual_reduction,
            gradient_norm,
            options_update_radius,
        )

    return (
        x_candidate,
        f_min,
        step_size_accepted,
        trustregion_radius,
        radius_lower_bound,
    )


def _update_trustregion_radius_conjugate_gradient(
    f_candidate,
    predicted_reduction,
    actual_reduction,
    x_norm_cg,
    trustregion_radius,
    options,
):
    """Update the trust-region radius based on predicted and actual reduction."""
    accept_step = False

    if predicted_reduction < 0 or ~np.isfinite(predicted_reduction):
        # Reject and start over
        trustregion_radius = options["alpha1"] * min(trustregion_radius, x_norm_cg)

    else:
        if ~np.isfinite(actual_reduction):
            trustregion_radius = options["alpha1"] * min(trustregion_radius, x_norm_cg)
        else:
            if abs(actual_reduction) <= max(1, abs(f_candidate) * EPSILON) and abs(
                predicted_reduction
            ) <= max(1, abs(f_candidate) * EPSILON):
                kappa = 1
            else:
                kappa = actual_reduction / predicted_reduction

            if kappa < options["eta1"]:
                # Reject the step
                trustregion_radius = options["alpha1"] * min(
                    trustregion_radius, x_norm_cg
                )
            else:
                accept_step = True

                # Update the trust-region radius only if the computed step is at the
                # trust-radius boundary
                if x_norm_cg == trustregion_radius:
                    if kappa < options["eta2"]:
                        # Marginal bad step
                        trustregion_radius = options["alpha2"] * trustregion_radius
                    elif kappa < options["eta3"]:
                        # Reasonable step
                        trustregion_radius = options["alpha3"] * trustregion_radius
                    elif kappa < options["eta4"]:
                        trustregion_radius = options["alpha4"] * trustregion_radius
                    else:
                        # Very good step
                        trustregion_radius = options["alpha5"] * trustregion_radius

    trustregion_radius = np.clip(
        trustregion_radius, options["min_radius"], options["max_radius"]
    )

    return trustregion_radius, accept_step


def _get_information_on_active_bounds(
    x,
    gradient_unprojected,
    lower_bounds,
    upper_bounds,
):
    """Return the index set of active bounds."""
    active_lower = np.where((x <= lower_bounds) & (gradient_unprojected > 0))[0]
    active_upper = np.where((x >= upper_bounds) & (gradient_unprojected < 0))[0]
    active_fixed = np.where(lower_bounds == upper_bounds)[0]
    active_all = reduce(np.union1d, (active_fixed, active_lower, active_upper))
    inactive = np.setdiff1d(np.arange(len(x)), active_all)

    active_bounds_info = ActiveBounds(
        lower=active_lower,
        upper=active_upper,
        fixed=active_fixed,
        active=active_all,
        inactive=inactive,
    )

    return active_bounds_info


def _find_hessian_submatrix_where_bounds_inactive(model, active_bounds_info):
    """Find the submatrix of the initial hessian where bounds are inactive."""
    hessian_inactive = model.square_terms[
        active_bounds_info.inactive[:, np.newaxis], active_bounds_info.inactive
    ]

    return hessian_inactive


def _check_for_convergence(
    x_candidate,
    f_candidate,
    gradient_candidate,
    model,
    lower_bounds,
    upper_bounds,
    converged,
    reason,
    niter,
    *,
    maxiter,
    gtol_abs,
    gtol_rel,
    gtol_scaled,
):
    """Check if we have found a solution."""
    direction_fischer_burmeister = _get_fischer_burmeister_direction_vector(
        x_candidate, gradient_candidate, lower_bounds, upper_bounds
    )
    gradient_norm = np.linalg.norm(direction_fischer_burmeister)
    gradient_norm_initial = np.linalg.norm(model.linear_terms)

    if gradient_norm < gtol_abs:
        converged = True
        reason = "Norm of the gradient is less than absolute_gradient_tolerance."
    elif f_candidate != 0 and abs(gradient_norm / f_candidate) < gtol_rel:
        converged = True
        reason = (
            "Norm of the gradient relative to the criterion value is less than "
            "relative_gradient_tolerance."
        )
    elif (
        gradient_norm_initial != 0
        and gradient_norm / gradient_norm_initial < gtol_scaled
    ):
        converged = True
        reason = (
            "Norm of the gradient divided by norm of the gradient at the "
            "initial parameters is less than scaled_gradient_tolerance."
        )
    elif gradient_norm_initial != 0 and gradient_norm == 0 and gtol_scaled == 0:
        converged = True
        reason = (
            "Norm of the gradient divided by norm of the gradient at the "
            "initial parameters is less than scaled_gradient_tolerance."
        )
    elif f_candidate <= -np.inf:
        converged = True
        reason = "Criterion value is negative infinity."
    elif niter is not None and niter == maxiter:
        reason = "Maximum number of iterations reached."

    return converged, reason


def _apply_bounds_to_x_candidate(x, lower_bounds, upper_bounds, bound_tol=0):
    """Apply upper and lower bounds to the candidate vector."""
    x = np.where(x <= lower_bounds + bound_tol, lower_bounds, x)
    x = np.where(x >= upper_bounds - bound_tol, upper_bounds, x)

    return x


def _project_gradient_onto_feasible_set(gradient_unprojected, active_bounds_info):
    """Project gradient onto feasible set, where search directions unconstrained."""
    gradient_projected = np.zeros_like(gradient_unprojected)
    gradient_projected[active_bounds_info.inactive] = gradient_unprojected[
        active_bounds_info.inactive
    ]

    return gradient_projected


def _apply_bounds_to_conjugate_gradient_step(
    step_inactive,
    x_candidate,
    lower_bounds,
    upper_bounds,
    active_bounds_info,
):
    """Apply lower and upper bounds to the Conjugate Gradient step."""
    cg_step = np.zeros_like(x_candidate)
    cg_step[active_bounds_info.inactive] = step_inactive

    if active_bounds_info.lower.size > 0:
        x_active_lower = x_candidate[active_bounds_info.lower]
        lower_bound_active = lower_bounds[active_bounds_info.lower]

        cg_step[active_bounds_info.lower] = lower_bound_active - x_active_lower

    if active_bounds_info.upper.size > 0:
        x_active_upper = x_candidate[active_bounds_info.upper]
        upper_bound_active = upper_bounds[active_bounds_info.upper]

        cg_step[active_bounds_info.upper] = upper_bound_active - x_active_upper

    if active_bounds_info.fixed.size > 0:
        cg_step[active_bounds_info.fixed] = 0

    return cg_step


def _update_trustregion_radius_and_gradient_descent(
    trustregion_radius,
    radius_lower_bound,
    predicted_reduction,
    actual_reduction,
    gradient_norm,
    options,
):
    """Update the trust-region radius and its upper bound."""
    if abs(actual_reduction) <= EPSILON and abs(predicted_reduction) <= EPSILON:
        kappa = 1
    else:
        kappa = actual_reduction / predicted_reduction

    tau_1 = (
        options["theta"]
        * gradient_norm
        * trustregion_radius
        / (
            options["theta"] * gradient_norm * trustregion_radius
            + (1 - options["theta"]) * predicted_reduction
            - actual_reduction
        )
    )
    tau_2 = (
        options["theta"]
        * gradient_norm
        * trustregion_radius
        / (
            options["theta"] * gradient_norm * trustregion_radius
            - (1 + options["theta"]) * predicted_reduction
            + actual_reduction
        )
    )

    tau_min = min(tau_1, tau_2)
    tau_max = max(tau_1, tau_2)

    if abs(kappa - 1) <= options["mu1"]:
        # Great agreement
        radius_lower_bound = max(radius_lower_bound, trustregion_radius)

        if tau_max < 1:
            tau = options["gamma3"]
        elif tau_max > options["gamma4"]:
            tau = options["gamma4"]
        else:
            tau = tau_max

    elif abs(kappa - 1) <= options["mu2"]:
        # Good agreement
        radius_lower_bound = max(radius_lower_bound, trustregion_radius)

        if tau_max < options["gamma2"]:
            tau = options["gamma2"]
        elif tau_max > options["gamma3"]:
            tau = options["gamma3"]
        else:
            tau = tau_max

    else:
        # Not good agreement
        if tau_min > 1:
            tau = options["gamma2"]
        elif tau_max < options["gamma1"]:
            tau = options["gamma1"]
        elif (tau_min < options["gamma1"]) and (tau_max >= 1):
            tau = options["gamma1"]
        elif (
            (tau_1 >= options["gamma1"])
            and (tau_1 < 1.0)
            and ((tau_2 < options["gamma1"]) or (tau_2 >= 1.0))
        ):
            tau = tau_1
        elif (
            (tau_2 >= options["gamma1"])
            and (tau_2 < 1.0)
            and ((tau_1 < options["gamma1"]) or (tau_2 >= 1.0))
        ):
            tau = tau_2
        else:
            tau = tau_max

    trustregion_radius = trustregion_radius * tau

    return trustregion_radius, radius_lower_bound


def _get_fischer_burmeister_direction_vector(x, gradient, lower_bounds, upper_bounds):
    """Compute the constrained direction vector via the Fischer-Burmeister function."""
    fischer_vec = np.vectorize(_get_fischer_burmeister_scalar)

    fischer_burmeister = reduce(
        fischer_vec, (upper_bounds - x, -gradient, x - lower_bounds)
    )
    direction = np.where(
        lower_bounds == upper_bounds, lower_bounds - x, fischer_burmeister
    )

    return direction


def _get_fischer_burmeister_scalar(a, b):
    """Get the value of the Fischer-Burmeister function for two scalar inputs.

    This method was suggested by Bob Vanderbei. Since the Fischer-Burmeister
    is symmetric, the order of the scalar inputs does not matter.

    Args:
        a (float): First input.
        b (float): Second input.

    Returns:
        float: Value of the Fischer-Burmeister function for inputs a and b.

    """
    if a + b <= 0:
        fischer_burmeister = np.sqrt(a**2 + b**2) - (a + b)
    else:
        fischer_burmeister = -2 * a * b / (np.sqrt(a**2 + b**2) + (a + b))

    return fischer_burmeister


def _evaluate_model_criterion(
    x,
    gradient,
    hessian,
):
    """Evaluate the criterion function value of the main model.

    Args:
        x (np.ndarray): Parameter vector of shape (n,).
        gradient (np.ndarray): Gradient of shape (n,) for which the main model
            shall be evaluated.
        hessian (np.ndarray): Hessian of shape (n, n) for which the main model
            shall be evaulated.

    Returns:
        float: Criterion value of the main model.

    """
    return gradient.T @ x + 0.5 * x.T @ hessian @ x
